
# _*_ coding: utf-8 _*_

'''
解析pup_rose文件

'''

# %%
import os
import numpy as np
import metradar.util.geo_transforms_pyart as geotrans
from datetime import datetime,timedelta
from .rose_structer import _unpack_from_buf,_structure_size
from .rose_structer import *

import xarray as xr
from pyart.core import Radar
from pyart.config import FileMetadata
from pyart.io.common import make_time_unit_str

class READ_ROSE(object):

    def __init__(self,):
        self.stiinfo = None
        self.mesoinfo = None
        self.hailinfo = None
        self.tvsinfo = None
        self.ssinfo = None
        pass
    
    # 将数字风暴序号转为字母数字组合

    def get_id_char(self,id_num):
        '''
        该函数的算法由张持岸提供
        '''
        if not isinstance(id_num,int):
            print('id_num is not int')
            return None
        if id_num > 0:
            num = (id_num-1) // 26
            tail = (id_num-1) % 26
            newid = '%s%d'%(chr(65+tail),num)
        else:
            newid = '0'
        return newid

    #将数字的风向转换为文字
    def get_wind_dir_name(self,wdir):
        wdir_cn = '未知'
        if not isinstance(wdir,int) and not isinstance(wdir,float):
            return wdir_cn
        if (wdir >= 348.76 and wdir <= 360) or (wdir >= 0 and wdir <= 11.25):
            wdir_cn = '北'
        elif wdir >= 11.26 and wdir <=33.75:
            wdir_cn = '北东北'
        elif wdir >= 33.76 and wdir <=56.25:
            wdir_cn = '东北'
        elif wdir >= 56.26 and wdir <=78.75:
            wdir_cn = '东东北'
        elif wdir >= 78.76 and wdir <101.25:
            wdir_cn = '东'
        elif wdir >= 101.26 and wdir <123.75:
            wdir_cn = '东东南'
        elif wdir >= 123.76 and wdir <146.25:
            wdir_cn = '东南'
        elif wdir >= 146.26 and wdir <168.75:
            wdir_cn = '南东南'
        elif wdir >= 168.76 and wdir <191.25:
            wdir_cn = '南'
        elif wdir >= 191.26 and wdir <213.75: 
            wdir_cn = '南西南'
        elif wdir >= 213.76 and wdir <236.25:
            wdir_cn = '西南'
        elif wdir >= 236.26 and wdir <258.75:
            wdir_cn = '西西南'
        elif wdir >= 258.76 and wdir <281.25:
            wdir_cn = '西'
        elif wdir >= 281.26 and wdir <303.75:
            wdir_cn = '西西北'
        elif wdir >= 303.76 and wdir <326.25:
            wdir_cn = '西北'
        elif wdir >= 326.26 and wdir <348.75:
            wdir_cn = '北西北'

        return wdir_cn
    
    def get_mda_rank(self,range_from_radar,shear_value):

        '''
        range_from_radar: km
        shear_value : m/s
        '''
        knot2ms = 0.514444445
        nmi     = 1.852 

        #x = np.array([0,30,60,90,120,150,180,210]) # x取值
        x  = np.arange(0,250) # x取值
        y1 = -1.*knot2ms*x/(24*nmi)+25*knot2ms
        y2 = -1.*knot2ms*x/(18*nmi)+35*knot2ms
        y3 = -1.*knot2ms*x/(14*nmi)+45*knot2ms

        curx = int(range_from_radar)
        if curx >=250 or curx <=0:
            print('range from radar should be larger than 0 and less than 250km ')
            return None
        
        cury = shear_value
        if cury < y1[curx]:
            return 1
        elif cury <y2[curx]:
            return 2
        elif cury < y3[curx]:
            return 3
        else:
            return 4

    def get_tvs_rank(self,lldv_value):

        # the "19" icon indicates that LLDV was at least 190 kts.

        ms2knot = 1.943844490

        value = lldv_value * ms2knot

        index = value // 10
        if index > 19:
            index=19

        return index+1
    
    # 解析风暴追踪产品
    def read_sti(self,filepath,filename):
        '''
        解析sti产品文件
        '''
        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        # pprint(prod_type)
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        pos = pos + 64 # 产品参数长度固定为64个


        # 获取产品数据块

        #获取STI头信息
        dic_sti_header = _unpack_from_buf(buf, pos, SIT_HEADER_BLOCK)
        pos = pos + _structure_size(SIT_HEADER_BLOCK)

        # 风暴追踪信息块
        # 风暴移动信息
        storm_motion_block=[]
        for nn in np.arange(dic_sti_header['storm_number']):
            storm_motion_block.append(_unpack_from_buf(buf, pos, STORM_MOTION_BLOCK))
            pos = pos + _structure_size(STORM_MOTION_BLOCK)

        # 风暴预报信息
        all_storm_fst_block=[]
        for nn in np.arange(dic_sti_header['storm_number']):
            dic_storm_fst_num = _unpack_from_buf(buf, pos, STORM_FST_HIS_NUM)
            pos = pos + _structure_size(STORM_FST_HIS_NUM)

            storm_fst_block=[]
            for nn in np.arange(dic_storm_fst_num['position_number']):
                storm_fst_block.append(_unpack_from_buf(buf, pos, STORM_FST_HIS_BLOCK))
                pos = pos + _structure_size(STORM_FST_HIS_BLOCK)
            all_storm_fst_block.append(storm_fst_block)

        # 风暴历史信息
        all_storm_his_block=[]
        for nn in np.arange(dic_sti_header['storm_number']):
            dic_storm_his_num = _unpack_from_buf(buf, pos, STORM_FST_HIS_NUM)
            pos = pos + _structure_size(STORM_FST_HIS_NUM)

            storm_his_block=[]
            for nn in np.arange(dic_storm_his_num['position_number']):
                storm_his_block.append(_unpack_from_buf(buf, pos, STORM_FST_HIS_BLOCK))
                pos = pos + _structure_size(STORM_FST_HIS_BLOCK)
            all_storm_his_block.append(storm_his_block)

        # 风暴属性表块数据
        # 风暴属性
        all_storm_prop=[]
        for nn in np.arange(dic_sti_header['storm_number']):
            dic_storm_prop = _unpack_from_buf(buf, pos, STORM_PROPERTY)
            pos = pos + _structure_size(STORM_PROPERTY)

            # 将风暴数字ID转换为字符串ID
            dic_storm_prop['id_char'] = self.get_id_char(dic_storm_prop['id'])

            # 将风暴移动信息融合进来
            dic_storm_prop['mv_spd'] = storm_motion_block[nn]['mv_spd']
            dic_storm_prop['mv_dir'] = storm_motion_block[nn]['mv_dir']

            all_storm_prop.append(dic_storm_prop)

        # 风暴构成表
        all_storm_comp=[]
        for nn in np.arange(dic_sti_header['storm_number']):
            dic_storm_comp = _unpack_from_buf(buf, pos, STORM_COMPONENT)
            pos = pos + _structure_size(STORM_COMPONENT)
            all_storm_comp.append(dic_storm_comp)

        # 风暴追踪适配数据
        dic_storm_track_param = _unpack_from_buf(buf, pos, STORM_TRACK_PARAM)
        pos = pos + _structure_size(STORM_TRACK_PARAM)


        allresult={}
        allresult['dic_gh'] = dic_gh
        allresult['dic_scfg'] = dic_scfg
        allresult['dic_tcfg'] = dic_tcfg
        allresult['cutinfo'] = cutinfo
        allresult['dic_prod_header'] = dic_prod_header
        allresult['dic_prod_param'] = dic_prod_param
        allresult['dic_sti_header'] = dic_sti_header
        allresult['storm_motion_block'] = storm_motion_block
        allresult['all_storm_fst_block'] = all_storm_fst_block
        allresult['all_storm_his_block'] = all_storm_his_block
        allresult['all_storm_prop'] = all_storm_prop
        allresult['all_storm_comp'] = all_storm_comp
        allresult['dic_storm_track_param'] = dic_storm_track_param

        # 将所有的方位角和距离转换成经纬度
        for nn in np.arange(dic_sti_header['storm_number']):
            pass
            x,y,z = geotrans.antenna_to_cartesian(storm_motion_block[nn]['range']/1000.0,storm_motion_block[nn]['azi'],0)
            clon,clat = geotrans.cartesian_to_geographic_aeqd(x,y,dic_scfg['lon'],dic_scfg['lat'])
            storm_motion_block[nn]['lon'] = clon[0]
            storm_motion_block[nn]['lat'] = clat[0]
            all_storm_prop[nn]['lon'] = clon[0]
            all_storm_prop[nn]['lat'] = clat[0]
            for blk in all_storm_fst_block[nn]:
                x,y,z = geotrans.antenna_to_cartesian(blk['range']/1000.0,blk['azi'],0)
                clon,clat = geotrans.cartesian_to_geographic_aeqd(x,y,dic_scfg['lon'],dic_scfg['lat'])
                blk['lon'] = clon[0]
                blk['lat'] = clat[0]
            for blk in all_storm_his_block[nn]:
                x,y,z = geotrans.antenna_to_cartesian(blk['range']/1000.0,blk['azi'],0)
                clon,clat = geotrans.cartesian_to_geographic_aeqd(x,y,dic_scfg['lon'],dic_scfg['lat'])
                blk['lon'] = clon[0]
                blk['lat'] = clat[0]

            pass

        # 整理成track
        alltrackinfo=[]
        all_past_position=[] # 历史位置
        all_fst_position=[] # 预报位置
        all_current_position=[] # 当前位置
        for nn in np.arange(dic_sti_header['storm_number']):
            pass
            
            if all_storm_prop[nn]['type']==0: # 连续风暴
                trackinfo=[]
                
                # 先添加历史位置
                for blk in all_storm_his_block[nn][::-1]:
                    trackinfo.append([blk['lat'],blk['lon']])
                    all_past_position.append([blk['lat'],blk['lon']])

                # 添加当前风暴位置
                trackinfo.append([storm_motion_block[nn]['lat'],storm_motion_block[nn]['lon']])
                all_current_position.append([storm_motion_block[nn]['lat'],storm_motion_block[nn]['lon']])

                # 再添加预报位置
                for blk in all_storm_fst_block[nn]:
                    trackinfo.append([blk['lat'],blk['lon']])
                    all_fst_position.append([blk['lat'],blk['lon']])

                alltrackinfo.append(trackinfo)

        # allresult={}
        allresult['track'] = alltrackinfo
        allresult['marker_past'] = all_past_position
        allresult['marker_current'] = all_current_position
        allresult['marker_fst'] = all_fst_position

        self.stiinfo = allresult
        return allresult

    # 解析风暴结构文本产品
    def read_ss(self,filepath,filename):
        '''
        解析ss产品文件
        '''
        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        # pprint(prod_type)
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        pos = pos + 64 # 产品参数长度固定为64个


        # 获取产品数据块

        #获取SS头信息
        dic_ss_header = _unpack_from_buf(buf, pos, SS_HEAD_BLOCK)
        pos = pos + _structure_size(SS_HEAD_BLOCK)

        # SS结构信息
        ss_tab=[]
        for nn in np.arange(dic_ss_header['storm_number']):
            tmp_tab = _unpack_from_buf(buf, pos, SS_TAB)
            pos = pos + _structure_size(SS_TAB)
            # 将风暴数字ID转换为字符串ID
            tmp_tab['id_char'] = self.get_id_char(tmp_tab['storm_id'])
            
            ss_tab.append(tmp_tab)


        # 风暴趋势信息
        cell_trend=[]
        for nn in np.arange(dic_ss_header['storm_number']):
            cur_cellinfo={}
            cell_info=_unpack_from_buf(buf, pos, CELL_TREND)
            pos = pos + _structure_size(CELL_TREND)
            cur_cellinfo['head_info'] = cell_info
            # 历史体扫信息
            tmpcell_his=[]
            for im in np.arange(cell_info['his_vol_num']):
                tmpcell_his.append(_unpack_from_buf(buf, pos, HIS_VOL))
                pos = pos + _structure_size(HIS_VOL)
            cur_cellinfo['cell_info'] = tmpcell_his

            cell_trend.append(cur_cellinfo)

        # 风暴段适配数据
        seg_adapt = _unpack_from_buf(buf, pos, SEG_ADAPT)
        pos = pos + _structure_size(SEG_ADAPT)

        # 风暴质心适配数据
        centroid_adapt = _unpack_from_buf(buf, pos, CENTROIDS_ADAPT)
        pos = pos + _structure_size(CENTROIDS_ADAPT)

        # 风暴追踪适配数据
        storm_track_adapt = _unpack_from_buf(buf, pos, STORM_TRACK_PARAM)
        pos = pos + _structure_size(STORM_TRACK_PARAM)
        

        # 将所有的方位角和距离转换成经纬度
        for nn in np.arange(dic_ss_header['storm_number']):
            pass
            x,y,z = geotrans.antenna_to_cartesian(ss_tab[nn]['range']/1000.0,ss_tab[nn]['azi'],0)
            clon,clat = geotrans.cartesian_to_geographic_aeqd(x,y,dic_scfg['lon'],dic_scfg['lat'])
            ss_tab[nn]['lon'] = clon[0]
            ss_tab[nn]['lat'] = clat[0]
        

        allresult={}
        allresult['ss'] = ss_tab
       
        self.ssinfo = allresult
        return allresult
    
    # 解析中气旋产品
    def read_mda(self,filepath,filename):
        '''
        解析meso产品文件
        '''
        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        # pprint(prod_type)
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        pos = pos + 64 # 产品参数长度固定为64个

        # 获取产品数据块

        #获取中气旋头信息
        dic_meso_header = _unpack_from_buf(buf, pos, MESO_HEADER_BLOCK)
        pos = pos + _structure_size(MESO_HEADER_BLOCK)

        # 中气旋表块
        meso_tab=[]
        for nn in np.arange(dic_meso_header['meso_number']):
            meso_tab.append(_unpack_from_buf(buf, pos, MESO_TABLE))
            pos = pos + _structure_size(MESO_TABLE)

        # 中气旋特征表
        meso_feature_tab=[]
        for nn in np.arange(dic_meso_header['feature_number']):
            meso_feature_tab.append(_unpack_from_buf(buf, pos, MESO_FEATURE_TAB))
            pos = pos + _structure_size(MESO_FEATURE_TAB)

        # 中气旋适配数据
        dic_meso_adapt_param = _unpack_from_buf(buf, pos, MESO_ADAPTATION_DATA)
        pos = pos + _structure_size(MESO_ADAPTATION_DATA)

        
        for nn in np.arange(dic_meso_header['meso_number']):
            # 将所有的方位角和距离转换成经纬度
            x,y,z = geotrans.antenna_to_cartesian(meso_tab[nn]['range']/1000.0,meso_tab[nn]['azi'],0)
            clon,clat = geotrans.cartesian_to_geographic_aeqd(x,y,dic_scfg['lon'],dic_scfg['lat'])
            meso_tab[nn]['lon'] = clon[0]
            meso_tab[nn]['lat'] = clat[0]
            
            # 将数字序号转换成字母数字组合
            meso_tab[nn]['storm_id_char'] = self.get_id_char(meso_tab[nn]['storm_id'])
            meso_tab[nn]['feature_id_char'] = self.get_id_char(meso_tab[nn]['feature_id'])
            meso_tab[nn]['feature_type'] = meso_feature_tab[nn]['feature_type']
            
        for nn in np.arange(dic_meso_header['feature_number']):
            # 将所有的方位角和距离转换成经纬度
            x,y,z = geotrans.antenna_to_cartesian(meso_feature_tab[nn]['range']/1000.0,meso_feature_tab[nn]['azi'],0)
            clon,clat = geotrans.cartesian_to_geographic_aeqd(x,y,dic_scfg['lon'],dic_scfg['lat'])
            meso_feature_tab[nn]['lon'] = clon[0]
            meso_feature_tab[nn]['lat'] = clat[0]
            
            # 将数字序号转换成字母数字组合
            meso_feature_tab[nn]['storm_id_char'] = self.get_id_char(meso_feature_tab[nn]['storm_id'])
            meso_feature_tab[nn]['feature_id_char'] = self.get_id_char(meso_feature_tab[nn]['feature_id'])


        allresult={}
        allresult['meso'] = meso_tab
        allresult['feature'] = meso_feature_tab
        
        self.mesoinfo = allresult
        return allresult

    # 解析龙卷涡旋特征产品
    def read_tvs(self,filepath,filename):
        '''
        解析tvs产品文件
        '''
        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        # pprint(prod_type)
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        pos = pos + 64 # 产品参数长度固定为64个

        # 获取产品数据块

        #获取TVS头信息
        dic_tvs_header = _unpack_from_buf(buf, pos, TVS_HEADER_BLOCK)
        pos = pos + _structure_size(TVS_HEADER_BLOCK)

        # TVS表块
        tvs_tab=[]
        for nn in np.arange(dic_tvs_header['tvs_number']):
            tvs_tab.append(_unpack_from_buf(buf, pos, TVS_TAB))
            pos = pos + _structure_size(TVS_TAB)

        # TVS适配数据
        dic_tvs_adapt_param = _unpack_from_buf(buf, pos, TVS_ADAPTATION_DATA)
        pos = pos + _structure_size(TVS_ADAPTATION_DATA)


        # 将所有的方位角和距离转换成经纬度
        for nn in np.arange(dic_tvs_header['tvs_number']):
            pass
            x,y,z = geotrans.antenna_to_cartesian(tvs_tab[nn]['range']/1000.0,tvs_tab[nn]['azi'],0)
            clon,clat = geotrans.cartesian_to_geographic_aeqd(x,y,dic_scfg['lon'],dic_scfg['lat'])
            tvs_tab[nn]['lon'] = clon[0]
            tvs_tab[nn]['lat'] = clat[0]
            
            
            # 将数字序号转换成字母数字组合
            tvs_tab[nn]['storm_id_char'] = self.get_id_char(tvs_tab[nn]['storm_id'])



        allresult={}
        allresult['tvs'] = tvs_tab
        
        self.tvsinfo = allresult
        return allresult
    
    # 解析ppi产品文件
    def read_ppi(self,filepath,filename):
        '''
        解析ppi产品文件
        '''
        pass

        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
            return None
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        pos = pos + 64 # 产品参数长度固定为64个

        # 读取产品径向头信息
        dic_radial_header = _unpack_from_buf(buf, pos, RADIAL_HEADER)
        pos = pos + _structure_size(RADIAL_HEADER)

        print('数据类型: %d-%s'%(dic_radial_header['data_type'],PRODUCT_DATA_TYPE[dic_radial_header['data_type']]))
        # Value=(Code-Offset)/Scale
        # 读取径向数据RADIAL_DATA
        data=[]
        data_azi=[]
        binnum = None
        for nn in np.arange(dic_radial_header['radial_number']):
            radial_data_info = _unpack_from_buf(buf, pos, RADIAL_DATA)
            pos = pos + _structure_size(RADIAL_DATA)
            binnum = radial_data_info['num_bins']
            rdata = buf[pos:pos+dic_radial_header['bin_length']*radial_data_info['num_bins']]
            data.extend(rdata)
            pos = pos + dic_radial_header['bin_length']*radial_data_info['num_bins']
            # print(radial_data_info['start_azi'])
            data_azi.append(radial_data_info['start_azi'])
            pass
        data = np.array(data)
       
        
        # data = data.astype('uint8')
        data = data.astype('float32')
        nx = dic_radial_header['radial_number']
        ny = binnum
        print(nx,ny)
        # 将data转成numpy数组
        data = data.reshape((nx,ny))
        
        
        # decode data
        # ppiarray = (ppiarray - dic_radial_header['offset']) / dic_radial_header['scale'] 

        # 将方位角坐标转换成经纬度坐标
        sweep_azimuths =np.array(data_azi)
        ele = dic_prod_param['ele']
        range_reso = dic_radial_header['resolution']
        # 输出的距离库只保留一半
        ngates = int(ny//2*2) #//2*2 
        grid_x = np.arange(-1*ngates,ngates+1,1)
        grid_y = np.arange(-1*ngates,ngates+1,1)
        total_gates = len(grid_x)

        aa = np.meshgrid(grid_x,grid_y)
        azi_grid = np.arctan2(aa[0],aa[1])*180/np.pi
        azi_grid[azi_grid<0]+=360

        azi_reso = 360/len(sweep_azimuths)

        # 求方位角索引
        new_azi = azi_grid.flatten()
        t = new_azi-sweep_azimuths[0]
        t[t<0]+=360
        ray_number = np.round(t/azi_reso,0).astype(int)
        ray_number[ray_number==len(sweep_azimuths)]=0
        ray_number = np.reshape(ray_number,(total_gates,total_gates))

        # 求距离索引
        dis_grid = np.sqrt(aa[0]**2 + aa[1]**2)
        dis_grid = np.round(dis_grid.flatten(),0).astype(int)
        dis_grid = np.reshape(dis_grid,(total_gates,total_gates))

        # 对数据进行截断，在径向方向上
        data = data[:,0:ngates]

        # data_grid = np.zeros((total_gates,total_gates),dtype='uint8') + 255
        data_grid = np.zeros((total_gates,total_gates),dtype='float32')
        new_data = data_grid.flatten()
        new_spdata = data.flatten()
        
        pos_out = [i+j*total_gates for i in range(total_gates) for j in range(total_gates) if dis_grid[i,j] < ngates]
        d_out = [rn*ngates+dg for rn,dg in zip(ray_number.flatten(),dis_grid.flatten()) if dg < ngates]

        new_data[pos_out]=new_spdata[d_out]
       
        data_grid = np.reshape(new_data,(total_gates,total_gates))
       

        if total_gates % 2 == 0:
            out_grid = np.arange(int(-total_gates/2),int(total_gates/2))
        else:
            out_grid = np.arange(int(-(total_gates-1)/2),int((total_gates-1)/2)+1)
        # 将ougrid转换为经纬度坐标
        out_lon,out_lat = geotrans.cartesian_to_geographic_aeqd(out_grid*range_reso,out_grid*range_reso,dic_scfg['lon'],dic_scfg['lat'])
        outdata = (data_grid - dic_radial_header['offset']) / dic_radial_header['scale'] 
        minvalue = (dic_radial_header['min_value'] - dic_radial_header['offset']) / dic_radial_header['scale']
        maxvalue = (dic_radial_header['max_value'] - dic_radial_header['offset']) / dic_radial_header['scale']
        # outdata[outdata > 90] = -32
        # outdata[outdata < -10] = -32
        outdata[outdata >= maxvalue] = -9999
        outdata[outdata <= minvalue] = -9999
        varname = 'ref'
        units = 'dBZ'
        if dic_radial_header['data_type'] == 7:
            varname = 'zdr'
            units = 'dB'
        elif dic_radial_header['data_type'] == 9:
            varname = 'cc'
            units = '%'
        elif dic_radial_header['data_type'] == 11:
            varname = 'kdp'
            units = 'deg/km'
            pass
        data = xr.DataArray(np.array(outdata.transpose()),coords=[out_lat,out_lon],dims=['lat','lon'],name=varname)
        data.attrs['units'] = units
        # data.attrs['standard_name'] = 'equivalent_reflectivity_factor'
        # data.attrs['long_name'] = 'equivalent_reflectivity_factor'
        data.attrs['varname'] = varname
        data.attrs['radar_lat'] = dic_scfg['lat']
        data.attrs['radar_lon'] = dic_scfg['lon']
        data.attrs['ana_height'] = dic_scfg['ana_height']
        data.attrs['grid_num'] = total_gates
        data.attrs['grid_reso'] = range_reso
        data.attrs['elevation'] = ele
        data.attrs['obs_range'] = int(range_reso * (total_gates-1)/2)
        data.attrs['distance_unit'] = 'meter'
        data.attrs['missing_value'] = -9999
        data.attrs['datatype'] = 'float32'
        # data.attrs['decode_method'] = 'dbz = (data - %d) / %d'%(dic_radial_header['offset'],dic_radial_header['scale'])
        data.attrs['task_name'] = dic_tcfg['task_name'].decode('utf-8').strip('\x00')
        data.attrs['radar_type'] = dic_scfg['radar_type']
        data.attrs['scan_time'] = datetime.fromtimestamp(dic_tcfg['scan_stime']).strftime('%Y-%m-%d %H:%M:%S')
        try:
            data.attrs['site_name'] = dic_scfg['site_name'].decode('utf-8').strip('\x00')
        except:
            data.attrs['site_name'] = 'Unknown'
        data.attrs['site_id'] = dic_scfg['site_code'].decode('utf-8').strip('\x00')
        # data.attrs['offset'] = dic_radial_header['offset']
        # data.attrs['scale'] = dic_radial_header['scale']

        return data

    # 解析cr产品文件，栅格数据
    def read_cr(self,filepath,filename):
        '''
        解析cr产品文件
        '''
        pass

        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
            return None
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        print('产品类型: %d-%s'%(dic_prod_header['product_type'],prod_type))
        pos = pos + 64 # 产品参数长度固定为64个

        # 读取产品径向头信息
        dic_grid_header = _unpack_from_buf(buf, pos, GRID_HEADER)
        pos = pos + _structure_size(GRID_HEADER)
        print('数据类型: %d-%s'%(dic_grid_header['data_type'],PRODUCT_DATA_TYPE[dic_grid_header['data_type']]))
        # Value=(Code-Offset)/Scale
        # 读取栅格数据GRID_DATA
        
        if dic_grid_header['bin_length'] == 1:
            data1 = np.frombuffer(buf[pos:pos + dic_grid_header['row_side_len']*dic_grid_header['col_side_len']*dic_grid_header['bin_length']], '>u1')
            data1 = data1.astype('uint8')
        data2 = np.reshape(data1,(dic_grid_header['row_side_len'],dic_grid_header['col_side_len']))
        # data2[data2<dic_grid_header['min_value']]=255
        # 根据offset和scale进行解码
        data2 = (data2 - dic_grid_header['offset']) / dic_grid_header['scale']
        data2[data2 > 90] = -32
        data2[data2 < -10] = -32
        nlat = dic_grid_header['row_side_len']
        nlon = dic_grid_header['col_side_len']
        # print(nlat,nlon)
        lat_reso = dic_grid_header['row_resolution']
        lon_reso = dic_grid_header['col_resolution']

        # 将data由list转成二维numpy数组

        
        # 将直角坐标转换成经纬度坐标
       
        if nlat % 2 == 0:
            out_grid_lat = np.arange(int(-nlat/2),int(nlat/2))
        else:
            out_grid_lat = np.arange(int(-(nlat-1)/2),int((nlat-1)/2)+1)
        
        if nlon % 2 == 0:
            out_grid_lon = np.arange(int(-nlon/2),int(nlon/2))
        else:
            out_grid_lon = np.arange(int(-(nlon-1)/2),int((nlon-1)/2)+1)

        # 将ougrid转换为经纬度坐标
        out_lon,out_lat = geotrans.cartesian_to_geographic_aeqd(out_grid_lat*lat_reso,out_grid_lon*lon_reso,dic_scfg['lon'],dic_scfg['lat'])

        data = xr.DataArray(np.flipud(data2),coords=[out_lat,out_lon],dims=['lat','lon'],name='cref')
        data.attrs['units'] = 'dBZ'
        data.attrs['standard_name'] = 'composite_reflectivity_factor'
        data.attrs['long_name'] = 'composite_reflectivity_factor'
        data.attrs['radar_lat'] = dic_scfg['lat']
        data.attrs['radar_lon'] = dic_scfg['lon']
        data.attrs['ana_height'] = dic_scfg['ana_height']
        data.attrs['lat_grid_num'] = nlat
        data.attrs['lon_grid_num'] = nlon
        data.attrs['lat_grid_reso'] = dic_grid_header['row_resolution']
        data.attrs['lon_grid_reso'] = dic_grid_header['col_resolution']
        data.attrs['distance_unit'] = 'meter'
        data.attrs['missing_value'] = -32
        data.attrs['datatype'] = 'float32'
        # data.attrs['decode_method'] = 'dbz = (data - %d) / %d'%(dic_grid_header['offset'],dic_grid_header['scale'])
        data.attrs['task_name'] = dic_tcfg['task_name'].decode('utf-8').strip('\x00')
        data.attrs['radar_type'] = dic_scfg['radar_type']
        data.attrs['scan_time'] = datetime.fromtimestamp(dic_tcfg['scan_stime']).strftime('%Y-%m-%d %H:%M:%S')
        try:
            data.attrs['site_name'] = dic_scfg['site_name'].decode('utf-8').strip('\x00')
        except:
            data.attrs['site_name'] = 'Unknown'
        data.attrs['site_id'] = dic_scfg['site_code'].decode('utf-8').strip('\x00')

        return data

    # 解析vil产品文件，栅格数据
    def read_vil(self,filepath,filename):
        '''
        解析vil产品文件
        '''
        pass

        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        print('产品类型: %d-%s'%(dic_prod_header['product_type'],prod_type))
        pos = pos + 64 # 产品参数长度固定为64个

        # 读取产品径向头信息
        dic_grid_header = _unpack_from_buf(buf, pos, GRID_HEADER)
        pos = pos + _structure_size(GRID_HEADER)
        print('数据类型: %d-%s'%(dic_grid_header['data_type'],PRODUCT_DATA_TYPE[dic_grid_header['data_type']]))
        # Value=(Code-Offset)/Scale
        # 读取栅格数据GRID_DATA
        
        if dic_grid_header['bin_length'] == 1:
            data1 = np.frombuffer(buf[pos:pos + dic_grid_header['row_side_len']*dic_grid_header['col_side_len']*dic_grid_header['bin_length']], '>u1')
            data1 = data1.astype('uint8')
        data2 = np.reshape(data1,(dic_grid_header['row_side_len'],dic_grid_header['col_side_len']))
        # data2[data2<dic_grid_header['min_value']]=255
        # 根据offset和scale进行解码
        data2 = (data2 - dic_grid_header['offset']) / dic_grid_header['scale']
        data2[data2 > 120] =  0
        data2[data2 < 0] = 0
        nlat = dic_grid_header['row_side_len']
        nlon = dic_grid_header['col_side_len']
        # print(nlat,nlon)
        lat_reso = dic_grid_header['row_resolution']
        lon_reso = dic_grid_header['col_resolution']

        # 将data由list转成二维numpy数组

        
        # 将直角坐标转换成经纬度坐标
       
        if nlat % 2 == 0:
            out_grid_lat = np.arange(int(-nlat/2),int(nlat/2))
        else:
            out_grid_lat = np.arange(int(-(nlat-1)/2),int((nlat-1)/2)+1)
        
        if nlon % 2 == 0:
            out_grid_lon = np.arange(int(-nlon/2),int(nlon/2))
        else:
            out_grid_lon = np.arange(int(-(nlon-1)/2),int((nlon-1)/2)+1)

        # 将ougrid转换为经纬度坐标
        out_lon,out_lat = geotrans.cartesian_to_geographic_aeqd(out_grid_lat*lat_reso,out_grid_lon*lon_reso,dic_scfg['lon'],dic_scfg['lat'])

        data = xr.DataArray(np.flipud(data2),coords=[out_lat,out_lon],dims=['lat','lon'],name='vil')
        data.attrs['units'] = 'kg/m^2'
        data.attrs['standard_name'] = 'vertically_integrated_liquid_water'
        data.attrs['long_name'] = 'vertically_integrated_liquid_water'
        data.attrs['radar_lat'] = dic_scfg['lat']
        data.attrs['radar_lon'] = dic_scfg['lon']
        data.attrs['ana_height'] = dic_scfg['ana_height']
        data.attrs['lat_grid_num'] = nlat
        data.attrs['lon_grid_num'] = nlon
        data.attrs['lat_grid_reso'] = dic_grid_header['row_resolution']
        data.attrs['lon_grid_reso'] = dic_grid_header['col_resolution']
        data.attrs['distance_unit'] = 'meter'
        data.attrs['missing_value'] = 0
        data.attrs['datatype'] = 'float32'
        # data.attrs['decode_method'] = 'dbz = (data - %d) / %d'%(dic_grid_header['offset'],dic_grid_header['scale'])
        data.attrs['task_name'] = dic_tcfg['task_name'].decode('utf-8').strip('\x00')
        data.attrs['radar_type'] = dic_scfg['radar_type']
        data.attrs['scan_time'] = datetime.fromtimestamp(dic_tcfg['scan_stime']).strftime('%Y-%m-%d %H:%M:%S')
        try:
            data.attrs['site_name'] = dic_scfg['site_name'].decode('utf-8').strip('\x00')
        except:
            data.attrs['site_name'] = 'Unknown'
        data.attrs['site_id'] = dic_scfg['site_code'].decode('utf-8').strip('\x00')

        return data
    
    # 解析tops产品文件，栅格数据，回波顶高
    def read_tops(self,filepath,filename):
        '''
        解析tops产品文件
        '''
        pass

        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        print('产品类型: %d-%s'%(dic_prod_header['product_type'],prod_type))
        pos = pos + 64 # 产品参数长度固定为64个

        # 读取产品径向头信息
        dic_grid_header = _unpack_from_buf(buf, pos, GRID_HEADER)
        pos = pos + _structure_size(GRID_HEADER)
        print('数据类型: %d-%s'%(dic_grid_header['data_type'],PRODUCT_DATA_TYPE[dic_grid_header['data_type']]))
        # Value=(Code-Offset)/Scale
        # 读取栅格数据GRID_DATA
        
        if dic_grid_header['bin_length'] == 1:
            data1 = np.frombuffer(buf[pos:pos + dic_grid_header['row_side_len']*dic_grid_header['col_side_len']*dic_grid_header['bin_length']], '>u1')
            data1 = data1.astype('uint8')
        data2 = np.reshape(data1,(dic_grid_header['row_side_len'],dic_grid_header['col_side_len']))
        # data2[data2<dic_grid_header['min_value']]=255
        # 根据offset和scale进行解码
        
        
        
        data2[data2 >  dic_grid_header['max_value']] =  0
        data2 = (data2 - dic_grid_header['offset']) / dic_grid_header['scale']
        maxv = (dic_grid_header['max_value'] - dic_grid_header['offset']) / dic_grid_header['scale']
        data2[data2 > maxv] = 0

        nlat = dic_grid_header['row_side_len']
        nlon = dic_grid_header['col_side_len']
        # print(nlat,nlon)
        lat_reso = dic_grid_header['row_resolution']
        lon_reso = dic_grid_header['col_resolution']

        # 将data由list转成二维numpy数组

        
        # 将直角坐标转换成经纬度坐标
       
        if nlat % 2 == 0:
            out_grid_lat = np.arange(int(-nlat/2),int(nlat/2))
        else:
            out_grid_lat = np.arange(int(-(nlat-1)/2),int((nlat-1)/2)+1)
        
        if nlon % 2 == 0:
            out_grid_lon = np.arange(int(-nlon/2),int(nlon/2))
        else:
            out_grid_lon = np.arange(int(-(nlon-1)/2),int((nlon-1)/2)+1)

        # 将ougrid转换为经纬度坐标
        out_lon,out_lat = geotrans.cartesian_to_geographic_aeqd(out_grid_lat*lat_reso,out_grid_lon*lon_reso,dic_scfg['lon'],dic_scfg['lat'])

        data = xr.DataArray(np.flipud(data2),coords=[out_lat,out_lon],dims=['lat','lon'],name='et')
        data.attrs['units'] = 'km'
        data.attrs['standard_name'] = 'echo_top'
        data.attrs['long_name'] = 'echo_top'
        data.attrs['radar_lat'] = dic_scfg['lat']
        data.attrs['radar_lon'] = dic_scfg['lon']
        data.attrs['ana_height'] = dic_scfg['ana_height']
        data.attrs['lat_grid_num'] = nlat
        data.attrs['lon_grid_num'] = nlon
        data.attrs['lat_grid_reso'] = dic_grid_header['row_resolution']
        data.attrs['lon_grid_reso'] = dic_grid_header['col_resolution']
        data.attrs['distance_unit'] = 'kilometer'
        data.attrs['missing_value'] = 0
        data.attrs['datatype'] = 'float32'
        # data.attrs['decode_method'] = 'dbz = (data - %d) / %d'%(dic_grid_header['offset'],dic_grid_header['scale'])
        data.attrs['task_name'] = dic_tcfg['task_name'].decode('utf-8').strip('\x00')
        data.attrs['radar_type'] = dic_scfg['radar_type']
        data.attrs['scan_time'] = datetime.fromtimestamp(dic_tcfg['scan_stime']).strftime('%Y-%m-%d %H:%M:%S')
        try:
            data.attrs['site_name'] = dic_scfg['site_name'].decode('utf-8').strip('\x00')
        except:
            data.attrs['site_name'] = 'Unknown'
        data.attrs['site_id'] = dic_scfg['site_code'].decode('utf-8').strip('\x00')

        return data
    

    # 解析stp产品文件，栅格数据
    def read_stp(self,filepath,filename):
        '''
        解析stp产品文件
        '''
        pass

        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        print('产品类型: %d-%s'%(dic_prod_header['product_type'],prod_type))
        pos = pos + 64 # 产品参数长度固定为64个

        # 读取产品径向头信息
        dic_radial_header = _unpack_from_buf(buf, pos, RADIAL_HEADER)
        pos = pos + _structure_size(RADIAL_HEADER)
        print('数据类型: %d-%s'%(dic_radial_header['data_type'],PRODUCT_DATA_TYPE[dic_radial_header['data_type']]))
        # Value=(Code-Offset)/Scale
        # 读取径向数据RADIAL_DATA
        data=[]
        data_azi=[]
        binnum = None
        for nn in np.arange(dic_radial_header['radial_number']):
            radial_data_info = _unpack_from_buf(buf, pos, RADIAL_DATA)
            pos = pos + _structure_size(RADIAL_DATA)
            binnum = radial_data_info['num_bins']
            # rdata = buf[pos:pos+dic_radial_header['bin_length']*radial_data_info['num_bins']]
            rdata = np.frombuffer(buf[pos:pos+dic_radial_header['bin_length']*radial_data_info['num_bins']], '<u2')
            data.extend(rdata)
            pos = pos + dic_radial_header['bin_length']*radial_data_info['num_bins']
            # print(radial_data_info['start_azi'])
            data_azi.append(radial_data_info['start_azi'])
            pass
        data = np.array(data)
        # data[data > dic_radial_header['max_value']] = 0
        # data[data < dic_radial_header['min_value']] = 0
        
        data = data.astype('uint16')
        nx = dic_radial_header['radial_number']
        ny = binnum
        print(nx,ny)
        # 将data转成numpy数组
        data = data.reshape((nx,ny))

        # 将方位角坐标转换成经纬度坐标
        sweep_azimuths =np.array(data_azi)

        range_reso = dic_radial_header['resolution']
        # 输出的距离库只保留一半
        ngates = int(ny//2*2) #//2*2 
        grid_x = np.arange(-1*ngates,ngates+1,1)
        grid_y = np.arange(-1*ngates,ngates+1,1)
        total_gates = len(grid_x)

        aa = np.meshgrid(grid_x,grid_y)
        azi_grid = np.arctan2(aa[0],aa[1])*180/np.pi
        azi_grid[azi_grid<0]+=360

        azi_reso = 360/len(sweep_azimuths)

        # 求方位角索引
        new_azi = azi_grid.flatten()
        t = new_azi-sweep_azimuths[0]
        t[t<0]+=360
        ray_number = np.round(t/azi_reso,0).astype(int)
        ray_number[ray_number==len(sweep_azimuths)]=0
        ray_number = np.reshape(ray_number,(total_gates,total_gates))

        # 求距离索引
        dis_grid = np.sqrt(aa[0]**2 + aa[1]**2)
        dis_grid = np.round(dis_grid.flatten(),0).astype(int)
        dis_grid = np.reshape(dis_grid,(total_gates,total_gates))

        # 对数据进行截断，在径向方向上
        data = data[:,0:ngates]

        data_grid = np.zeros((total_gates,total_gates),dtype='uint16') + 0
        new_data = data_grid.flatten()
        new_spdata = data.flatten()
        # decode data
        new_spdata = (new_spdata - dic_radial_header['offset']) / dic_radial_header['scale'] 
        pos_out = [i+j*total_gates for i in range(total_gates) for j in range(total_gates) if dis_grid[i,j] < ngates]
        d_out = [rn*ngates+dg for rn,dg in zip(ray_number.flatten(),dis_grid.flatten()) if dg < ngates]

        new_data[pos_out]=new_spdata[d_out]
       
        data_grid = np.reshape(new_data,(total_gates,total_gates))


        if total_gates % 2 == 0:
            out_grid = np.arange(int(-total_gates/2),int(total_gates/2))
        else:
            out_grid = np.arange(int(-(total_gates-1)/2),int((total_gates-1)/2)+1)
        # 将ougrid转换为经纬度坐标
        out_lon,out_lat = geotrans.cartesian_to_geographic_aeqd(out_grid*range_reso,out_grid*range_reso,dic_scfg['lon'],dic_scfg['lat'])

        data = xr.DataArray(np.array(data_grid.transpose()),coords=[out_lat,out_lon],dims=['lat','lon'],name='stp')
        data.attrs['units'] = 'mm'
        data.attrs['standard_name'] = 'storm total precipitation'
        data.attrs['long_name'] = 'storm total precipitation'
        data.attrs['radar_lat'] = dic_scfg['lat']
        data.attrs['radar_lon'] = dic_scfg['lon']
        data.attrs['ana_height'] = dic_scfg['ana_height']
        data.attrs['grid_num'] = total_gates
        data.attrs['grid_reso'] = range_reso
        data.attrs['obs_range'] = int(range_reso * (total_gates-1)/2)
        data.attrs['distance_unit'] = 'meter'
        data.attrs['missing_value'] = 0
        data.attrs['datatype'] = 'uint16'
        # data.attrs['decode_method'] = 'qpe = (data - %d) / %d'%(dic_radial_header['offset'],dic_radial_header['scale'])
        data.attrs['task_name'] = dic_tcfg['task_name'].decode('utf-8').strip('\x00')
        data.attrs['radar_type'] = dic_scfg['radar_type']
        data.attrs['scan_time'] = datetime.fromtimestamp(dic_tcfg['scan_stime']).strftime('%Y-%m-%d %H:%M:%S')
        try:
            data.attrs['site_name'] = dic_scfg['site_name'].decode('utf-8').strip('\x00')
        except:
            data.attrs['site_name'] = 'Unknown'
        data.attrs['site_id'] = dic_scfg['site_code'].decode('utf-8').strip('\x00')
        # data.attrs['offset'] = dic_radial_header['offset']
        # data.attrs['scale'] = dic_radial_header['scale']

        return data

    # 解析ohp产品文件，栅格数据
    def read_ohp(self,filepath,filename):
        '''
        解析ohp产品文件
        '''
        pass

        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
            return None
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        print('产品类型: %d-%s'%(dic_prod_header['product_type'],prod_type))
        pos = pos + 64 # 产品参数长度固定为64个

        # 读取产品径向头信息
        dic_radial_header = _unpack_from_buf(buf, pos, RADIAL_HEADER)
        pos = pos + _structure_size(RADIAL_HEADER)
        print('数据类型: %d-%s'%(dic_radial_header['data_type'],PRODUCT_DATA_TYPE[dic_radial_header['data_type']]))
        # Value=(Code-Offset)/Scale
        # 读取径向数据RADIAL_DATA
        data=[]
        data_azi=[]
        binnum = None
        for nn in np.arange(dic_radial_header['radial_number']):
            radial_data_info = _unpack_from_buf(buf, pos, RADIAL_DATA)
            pos = pos + _structure_size(RADIAL_DATA)
            binnum = radial_data_info['num_bins']
            # rdata = buf[pos:pos+dic_radial_header['bin_length']*radial_data_info['num_bins']]
            rdata = np.frombuffer(buf[pos:pos+dic_radial_header['bin_length']*radial_data_info['num_bins']], '<u2')
            data.extend(rdata)
            pos = pos + dic_radial_header['bin_length']*radial_data_info['num_bins']
            # print(radial_data_info['start_azi'])
            data_azi.append(radial_data_info['start_azi'])
            pass
        data = np.array(data)
        data[data > dic_radial_header['max_value']] = 0
        data[data < dic_radial_header['min_value']] = 0
        
        data = data.astype('uint16')
        nx = dic_radial_header['radial_number']
        ny = binnum
        print(nx,ny)
        # 将data转成numpy数组
        data = data.reshape((nx,ny))

        # decode data
        # ppiarray = (ppiarray - dic_radial_header['offset']) / dic_radial_header['scale'] 

        
        # 将方位角坐标转换成经纬度坐标
        sweep_azimuths =np.array(data_azi)

        range_reso = dic_radial_header['resolution']
        # 输出的距离库只保留一半
        ngates = int(ny//2*2) #//2*2 
        grid_x = np.arange(-1*ngates,ngates+1,1)
        grid_y = np.arange(-1*ngates,ngates+1,1)
        total_gates = len(grid_x)

        aa = np.meshgrid(grid_x,grid_y)
        azi_grid = np.arctan2(aa[0],aa[1])*180/np.pi
        azi_grid[azi_grid<0]+=360

        azi_reso = 360/len(sweep_azimuths)

        # 求方位角索引
        new_azi = azi_grid.flatten()
        t = new_azi-sweep_azimuths[0]
        t[t<0]+=360
        ray_number = np.round(t/azi_reso,0).astype(int)
        ray_number[ray_number==len(sweep_azimuths)]=0
        ray_number = np.reshape(ray_number,(total_gates,total_gates))

        # 求距离索引
        dis_grid = np.sqrt(aa[0]**2 + aa[1]**2)
        dis_grid = np.round(dis_grid.flatten(),0).astype(int)
        dis_grid = np.reshape(dis_grid,(total_gates,total_gates))

        # 对数据进行截断，在径向方向上
        data = data[:,0:ngates]

        data_grid = np.zeros((total_gates,total_gates),dtype='uint16') + 0
        new_data = data_grid.flatten()
        new_spdata = data.flatten()

        new_spdata = (new_spdata - dic_radial_header['offset']) / dic_radial_header['scale'] 

        pos_out = [i+j*total_gates for i in range(total_gates) for j in range(total_gates) if dis_grid[i,j] < ngates]
        d_out = [rn*ngates+dg for rn,dg in zip(ray_number.flatten(),dis_grid.flatten()) if dg < ngates]

        new_data[pos_out]=new_spdata[d_out]
       
        data_grid = np.reshape(new_data,(total_gates,total_gates))
       

        if total_gates % 2 == 0:
            out_grid = np.arange(int(-total_gates/2),int(total_gates/2))
        else:
            out_grid = np.arange(int(-(total_gates-1)/2),int((total_gates-1)/2)+1)
        # 将ougrid转换为经纬度坐标
        out_lon,out_lat = geotrans.cartesian_to_geographic_aeqd(out_grid*range_reso,out_grid*range_reso,dic_scfg['lon'],dic_scfg['lat'])

        data = xr.DataArray(np.array(data_grid.transpose()),coords=[out_lat,out_lon],dims=['lat','lon'],name='ohp')
        data.attrs['units'] = 'mm'
        data.attrs['standard_name'] = 'one hour precipitation'
        data.attrs['long_name'] = 'one hour precipitation'
        data.attrs['radar_lat'] = dic_scfg['lat']
        data.attrs['radar_lon'] = dic_scfg['lon']
        data.attrs['ana_height'] = dic_scfg['ana_height']
        data.attrs['grid_num'] = total_gates
        data.attrs['grid_reso'] = range_reso
        data.attrs['obs_range'] = int(range_reso * (total_gates-1)/2)
        data.attrs['distance_unit'] = 'meter'
        data.attrs['missing_value'] = 0
        data.attrs['datatype'] = 'uint16'
        # data.attrs['decode_method'] = 'qpe = (data - %d) / %d'%(dic_radial_header['offset'],dic_radial_header['scale'])
        data.attrs['task_name'] = dic_tcfg['task_name'].decode('utf-8').strip('\x00')
        data.attrs['radar_type'] = dic_scfg['radar_type']
        data.attrs['scan_time'] = datetime.fromtimestamp(dic_tcfg['scan_stime']).strftime('%Y-%m-%d %H:%M:%S')
        try:
            data.attrs['site_name'] = dic_scfg['site_name'].decode('utf-8').strip('\x00')
        except:
            data.attrs['site_name'] = 'Unknown'
        data.attrs['site_id'] = dic_scfg['site_code'].decode('utf-8').strip('\x00')
        # data.attrs['offset'] = dic_radial_header['offset']
        # data.attrs['scale'] = dic_radial_header['scale']

        return data

    
    # 解析冰雹指数产品
    def read_hda(self,filepath,filename):
        '''
        解析hail产品文件
        '''
        fin = open(filepath + os.sep + filename,'rb')
        buf = fin.read()
        fin.close()
        buf_length = len(buf)
        pos = 0

        # 获取通用信息头
        dic_gh = _unpack_from_buf(buf, pos, GENERIC_HEADER)
        # pprint(dic_gh)
        if dic_gh['magic_number'] != 0x4D545352:
            print('源数据格式错误！')
        pos = pos + _structure_size(GENERIC_HEADER)

        # 获取站点信息
        dic_scfg = _unpack_from_buf(buf, pos, SITE_CONFIG)
        pos = pos + _structure_size(SITE_CONFIG)

        # 获取任务信息
        dic_tcfg = _unpack_from_buf(buf, pos, TASK_CONFIG)
        pos = pos + _structure_size(TASK_CONFIG)

        # 获取扫描信息
        cutinfo = []
        for im in np.arange(dic_tcfg['cut_number']):
            dic_cutcfg = _unpack_from_buf(buf, pos, SCAN_CONFIG)
            cutinfo.append(dic_cutcfg)
            pos = pos + _structure_size(SCAN_CONFIG)

        # 获取产品头信息
        dic_prod_header = _unpack_from_buf(buf, pos, PRODUCT_HEADER_BLOCK)
        pos = pos + _structure_size(PRODUCT_HEADER_BLOCK)

        # 获取产品参数信息
        prod_type = PRODUCT_ID_NAME_MAP[dic_prod_header['product_type']]
        # pprint(prod_type)
        dic_prod_param = _unpack_from_buf(buf, pos, PRODUCT_DEPENDENT_PARAMETER[prod_type])
        pos = pos + 64 # 产品参数长度固定为64个

        # 获取产品数据块
        # 冰雹个数
        hail_number = _unpack_from_buf(buf, pos, HAIL_NUMBER)
        pos = pos + _structure_size(HAIL_NUMBER)

        # TVS表块
        hail_tab=[]
        for nn in np.arange(hail_number['hail_number']):
            hail_tab.append(_unpack_from_buf(buf, pos, HAIL_TAB))
            pos = pos + _structure_size(HAIL_TAB)

        # TVS适配数据
        dic_hail_adapt_param = _unpack_from_buf(buf, pos, HAIL_ADAPTATION_DATA)
        pos = pos + _structure_size(HAIL_ADAPTATION_DATA)

        # 将所有的方位角和距离转换成经纬度
        for nn in np.arange(hail_number['hail_number']):
            pass
            x,y,z = geotrans.antenna_to_cartesian(hail_tab[nn]['range']/1000.0,hail_tab[nn]['azi'],0)
            clon,clat = geotrans.cartesian_to_geographic_aeqd(x,y,dic_scfg['lon'],dic_scfg['lat'])
            hail_tab[nn]['lon'] = clon[0]
            hail_tab[nn]['lat'] = clat[0]

            # 将数字序号转换成字母数字组合
            hail_tab[nn]['hail_id_char'] = self.get_id_char(hail_tab[nn]['hail_id'])


        allresult={}
        allresult['hail'] = hail_tab

        self.hailinfo = allresult
        return allresult
    
        
if __name__ == "__main__":
    rootpath = 'testdata/rose_product'
    staid = 'Z9852'

    decoder = READ_ROSE()

    # # ppi zdr
    # filepath_ppi = '/Users/wenjianzhu/Downloads/CMADAAS/pup/Z9852/SS'

    # # ss
    # filepath_ss = rootpath + os.sep + staid + os.sep + 'SS'
    # filename_ss = 'Z_RADR_I_Z9852_20230421164954_P_DOR_CDD_SS_NUL_300_NUL_FMT.bin'
    # allresult_ss = decoder.read_ss(filepath_ss, filename_ss)
    

    # sti
    # filepath_sti = rootpath + os.sep + staid + os.sep + 'STI'
    # filename_sti = 'Z_RADR_I_Z9852_20230421164954_P_DOR_CDD_STI_NUL_300_NUL_FMT.bin'
    # allresult_sti = decoder.read_sti(filepath_sti, filename_sti)
    
    
    # # meso
    # filepath_meso = rootpath + os.sep + staid + os.sep + 'M'
    # filename_meso = 'Z_RADR_I_Z9852_20230421135042_P_DOR_CDD_M_NUL_200_NUL_FMT.bin'
    # allresult_meso = decoder.read_mda(filepath_meso, filename_meso)
   

    # # tvs
    # filepath_tvs = rootpath + os.sep + staid + os.sep + 'TVS'
    # filename_tvs = 'Z_RADR_I_Z9852_20230421201451_P_DOR_CDD_TVS_NUL_200_NUL_FMT.bin'
    # allresult_tvs = decoder.read_tvs(filepath_tvs, filename_tvs)
   


    # # hail
    # filepath_hail = rootpath + os.sep + staid + os.sep + 'HI'
    # filename_hail = 'Z_RADR_I_Z9852_20230421181642_P_DOR_CDD_HI_NUL_200_NUL_FMT.bin'
    # allresult_hail = decoder.read_hda(filepath_hail, filename_hail)
   

# %%



# _*_ coding: utf-8 _*_

'''
从pyart的radar object中获取任意垂直剖面的数据

朱文剑


'''

import numpy as np
from pyart.core import Radar,antenna_to_cartesian
import math
# import time

R = 6371 # km
#该程序用来求直线方程系数
#目前没考虑垂直X轴的直线，需要完善
#kevin2075@163.com
def linefunc(x,y,x1,y1): #直接输入两个坐标就可以求出一条直线方程。
# if x~=x1
    a=(y1-y)/(x1-x)
    b=y-a*x
    return [a,b]


# 将多个数或列表组合成一个列表
def myconcat(vars:tuple):
    result=[]
    for vartmp in vars:
        if len(np.array(vartmp).shape)==0:
            result.append(vartmp)
        else:
            for var in vartmp:
                result.append(var)
   
    return result

def get_cross_radar(radar:Radar,params):

    # antenna_to_cartesian(147, 260, 1.5)
    # startangle=None,startrange=None,endangle=None,endrange=None
    # startangle = 8.5
    # startrange = 69.23 #km

    # endangle = 18.6;
    # endrange =  59.03 #km
    startangle = params['crs_start_azi']
    startrange = params['crs_start_range']
    endangle = params['crs_end_azi']
    endrange = params['crs_end_range']
    toph = params['top_height']

    startx = startrange*np.sin(np.radians(startangle))
    starty = startrange*np.cos(np.radians(startangle))
    endx   = endrange*np.sin(np.radians(endangle))
    endy   = endrange*np.cos(np.radians(endangle))

    baseh = 0; # km 垂直剖面的起始高度
    # toph = 20; # km 垂直剖面的上限高度

    validzlv=[]
    validvlv=[]
    validReles=[]
    validDeles=[]
    outdic=dict()

    for il in range(radar.nsweeps):
        # if sweep(il).RGates > 0
        # if il !=1 and il !=3:
        #     validzlv.append(il)
        #     validReles.append(radar.get_elevation(il).mean())
        # if il !=0 and il !=2:
        #     validvlv.append(il)
        #     validDeles.append(radar.get_elevation(il).mean())
        validzlv.append(il)
        validReles.append(round(radar.get_elevation(il).mean(),1))
        validvlv.append(il)
        validDeles.append(round(radar.get_elevation(il).mean(),1))

    tol_length = int(np.sqrt(np.power((endx-startx),2) + np.power((endy - starty),2)))

    #求方程坐标，这里暂时没考虑X=0的情况，后面要完善这一点
    [a,b] = linefunc(startx,starty,endx,endy)

    # 获取反射率垂直剖面
        
    #构建变量，确定坐标维度
    xreso = 0.05#km
    if endx < startx:
        xreso = -1*xreso

    xnum = int(tol_length  / abs(xreso))
    zreso = 0.05# km
    znum = int((toph - baseh)  / zreso)

    

    beamwidth = radar.get_azimuth(validzlv[-1])
    beamwidth = np.diff(beamwidth)
    beamwidth = np.mean(beamwidth[beamwidth>0])
    topel_R = max(validReles) + beamwidth / 2
    baseel_R = min(validReles) - beamwidth / 2

    topel_D = max(validDeles) + beamwidth / 2
    baseel_D = min(validDeles) - beamwidth / 2

    tmp=[]
    for nn in range(len(validReles)-1):
        tmp.append(np.mean([validReles[nn],validReles[nn+1]]))

    tmp_newRel = sorted(myconcat((tmp,validReles)))

    tmp=[]
    for nn in range(len(validDeles)-1):
        tmp.append(np.mean([validDeles[nn],validDeles[nn+1]]))

    tmp_newDel = sorted(tmp+validDeles)

    newRel = [baseel_R+tmp_newRel+topel_R]
    newDel = [baseel_D+tmp_newDel+topel_D]

    # for varkey in radar.fields.keys():
     
    nflag=0
    outdic=dict()
    outdic['vertical_km'] = toph - baseh
    outdic['horizontal_km'] = tol_length
    outdic['xreso'] = xreso
    outdic['yreso'] = zreso
    
    for varkey in radar.fields.keys():
        outdic[varkey]=[]
        outdic[varkey] = np.zeros([znum,xnum],dtype='float')*np.nan

    
    # 纯 for 循环版本 =====================
    valideles=[]
    validlvl=[]
    for varkey in radar.fields.keys():
        print(varkey)
        if varkey == 'reflectivity' :
            valideles = validReles
            validlvl = validzlv
        elif varkey == 'velocity':
            valideles = validDeles
            validlvl = validvlv 
        else:
            # print('暂时仅支持回波强度和径向速度')
            continue
        
        curfd = []
        for il in range(radar.nsweeps):
            curfd.append(radar.get_field(il,varkey))#sweep(targetsp).dbz(az,gates);

        for iz in range(znum):
            for ix in range(xnum):
                tmpcurx = startx + (ix * xreso) * math.cos(math.atan(a))
                
                tmpcury = a * tmpcurx + b
                tmpaz = (90 - math.degrees(math.atan2(tmpcury,tmpcurx)) + 360)%360
                
                curx = math.sqrt(math.pow(tmpcurx,2) + math.pow(tmpcury,2))#startrange + xreso * ix 
                curz = baseh + iz * zreso
                
                #计算相对雷达的仰角
                # ele = math.degrees(math.atan2(curz,curx))
                # 这里必须用math.pow，如果直接用称号会有较大误差
                # print('R=',R,' curz=',curz,' curx=',curx)
                ele = math.asin((math.pow((curz+R),2) - math.pow(curx,2) - math.pow(R,2))/2/curx/R)/np.pi*180

                if ele < baseel_R or ele > topel_R:
                    continue

                tmp = list(abs(np.array(valideles) - ele))
                idx = tmp.index(min(tmp))
                idx_el = idx
                
                targetsp = validlvl[idx_el]
                
                #获取方位角角标
                tmp = list(abs(radar.get_azimuth(targetsp) - tmpaz))
                idx = tmp.index(min(tmp))
                az = idx
                
                #经向距离，由该页面的公式进行反算
                #https://arm-doe.github.io/pyart/API/generated/pyart.core.antenna_to_cartesian.html#pyart.core.antenna_to_cartesian
                #根据s=R*??的公式反算r
                
                cur_range = (R+curz)*math.sin(curx/R)/math.cos(math.radians(valideles[idx_el]))
                gates = math.floor(cur_range/(radar.range['meters_between_gates'] / 1000)+0.5)-1
                # gates = math.floor(curx/math.cos(math.radians(valideles[idx_el]))/(radar.range['meters_between_gates'] / 1000)+0.5)-1
                
                if gates < radar.ngates:
                    #disp(sweep(targetsp).dbz(az,gates));
                    # if curfd[targetsp].data[az,gates] > -30 and not curfd[targetsp].mask[az,gates]:
                    # if curfd[targetsp].data[az,gates] > -5:
                    if not curfd[targetsp].mask[az,gates]:
                        outdic[varkey][iz,ix] = curfd[targetsp].data[az,gates]
                        nflag+=1
    
    return outdic



# _*_ coding: utf-8 _*_

# Copyright (c) 2026 NMC Developers.
# Distributed under the terms of the GPL V3 License.

"""
Read configure file.
"""



import os
import configparser
from pathlib import Path


def _get_config_dir():
    """
    Get default configuration directory.
    """
    config_dir = Path.home() / ".metradar"
    config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir

# Global Variables
CONFIG_DIR = _get_config_dir()


def _ConfigFetchError(BaseException):
    pass


def _get_config_from_rcfile(rc='config.ini'):
    """
    Get configure information from config_dk_met_io.ini file.
    """

    if not os.path.exists(rc):
        print(str(rc) + os.sep + ' not exists!')
        return None
  
    try:
        config = configparser.ConfigParser()
        config.read(rc,encoding='utf-8')
    except IOError as e:
        raise _ConfigFetchError(str(e))
    except BaseException as e:
        raise _ConfigFetchError(str(e))

    return config

# Global Variables
CONFIG = _get_config_from_rcfile(CONFIG_DIR / 'config.ini')




# _*_ coding: utf-8 _*_

'''
读取swan格点数据

'''

# %%
import numpy as np
import xarray as xr
import bz2
import os
from datetime import datetime, timedelta
from metradar.graph.draw_latlon_func import draw_latlon


# define head structure
head_dtype = [
    ('ZonName', 'S12'),
    ('DataName', 'S38'),
    ('Flag', 'S8'),
    ('Version', 'S8'),
    ('year', 'i2'),
    ('month', 'i2'),     
    ('day', 'i2'),     
    ('hour', 'i2'),    
    ('minute', 'i2'),
    ('interval', 'i2'),
    ('XNumGrids', 'i2'),
    ('YNumGrids', 'i2'),
    ('ZNumGrids', 'i2'),
    ('RadarCount', 'i4'),
    ('StartLon', 'f4'),
    ('StartLat', 'f4'),
    ('CenterLon', 'f4'),
    ('CenterLat', 'f4'),
    ('XReso', 'f4'),
    ('YReso', 'f4'),
    ('ZhighGrids', 'f4', 40),
    ('RadarStationName', 'S20', 16),
    ('RadarLongitude', 'f4', 20),
    ('RadarLatitude', 'f4', 20),
    ('RadarAltitude', 'f4', 20),
    ('MosaicFlag', 'S1', 20),
    ('m_iDataType', 'i2'),
    ('m_iLevelDimension', 'i2'),
    ('offset','f4'),
    ('scale','f4'),
    ('Reserved', 'S160')]


def decode_swan(filepath,filename,scale=None,varattrs={'long_name': 'mosaic composite reflectivity', 'short_name': 'QPF', 'units': 'mm'}
                ,attach_forecast_period=False):

    if not os.path.exists(filepath+os.sep+filename):
        print(filepath + os.sep + filename + ' not exists!')
        return None
        
    fp = open(filepath + os.sep + filename, 'rb')

    if filename.find('.bz2') > 0:
        
        tmpbuf = fp.read()
        byteArray = bz2.decompress(tmpbuf)
    else:
        byteArray = fp.read()

    # read head information
    head_info = np.frombuffer(byteArray[0:1024], dtype=head_dtype)
    ind = 1024

    # get coordinates
    nlon = head_info['XNumGrids'][0].astype(np.int64)
    nlat = head_info['YNumGrids'][0].astype(np.int64)
    nlev = head_info['ZNumGrids'][0].astype(np.int64)
    dlon = head_info['XReso'][0].astype(np.float32)
    dlat = head_info['YReso'][0].astype(np.float32)
    lat = head_info['StartLat'][0] - np.arange(nlat)*dlat - dlat/2.0
    lon = head_info['StartLon'][0] + np.arange(nlon)*dlon - dlon/2.0
    level = head_info['ZhighGrids'][0][0:nlev]
    cur_scale = head_info['scale'][0].astype(np.float32)
    cur_offset = head_info['offset'][0].astype(np.float32)

    # retrieve data records
    data_type = ['u1', 'u1', 'u2', 'i2']
    data_type = data_type[head_info['m_iDataType'][0]]
    data_len = (nlon * nlat * nlev)
    data = np.frombuffer(
        byteArray[ind:(ind + data_len*int(data_type[1]))], 
        dtype=data_type, count=data_len)

    # convert data type
    data.shape = (nlev, nlat, nlon)
    data = data.astype(np.float32)
    if not scale is None:
        data = (data + scale[1]) * scale[0]
    else:
        data = data * cur_scale + cur_offset

    # reverse latitude axis
    # if len(data.shape) == 3:
    #     data = np.flip(data, 2)
    # else:
    data = np.flip(data, 1)
    lat = lat[::-1]

    # set time coordinates
    init_time = datetime(
        head_info['year'][0], head_info['month'][0], 
        head_info['day'][0], head_info['hour'][0], head_info['minute'][0])
    if attach_forecast_period:
        fhour = int(filename.split('.')[1])/60.0
    else:
        fhour = 0
    fhour = np.array([fhour], dtype=np.float64)
    time = init_time + timedelta(hours=fhour[0])
    init_time = np.array([init_time], dtype='datetime64[ns]')
    time = np.array([time], dtype='datetime64[ns]')

    # define coordinates
    time_coord = ('time', time)
    lon_coord = ('lon', lon, {
        'long_name':'longitude', 'units':'degrees_east', '_CoordinateAxisType':'Lon'})
    lat_coord = ('lat', lat, {
        'long_name':'latitude', 'units':'degrees_north', '_CoordinateAxisType':'Lat'})
    level_coord = ('level', level, {
        'long_name':'height', 'units':'m'})

    # create xarray
    data = np.expand_dims(data, axis=0)
    data = xr.Dataset({'data':(['time', 'level', 'lat', 'lon'], data, varattrs)},
        coords={'time':time_coord, 'level':level_coord, 'lat':lat_coord, 'lon':lon_coord})

    # add time coordinates
    data.coords['forecast_reference_time'] = init_time[0]
    data.coords['forecast_period'] = ('time', fhour, {
        'long_name':'forecast_period', 'units':'hour'})

    # add attributes
    data.attrs['Conventions'] = "CF-1.6"

    return data



if __name__ == "__main__":
#     pass

# %%
    # testdata/SWAN/

    # prod_type = 'MCR'
    prod_type = 'MCC'
    # prod_type = 'MKDP'
    # prod_type = 'MZDR'
    # prod_type = 'HCAMOSAIC'
    # prod_type = 'TOP'
    # prod_type = 'VIL'

    timestr = '20230316113000'
    filepath = '/home/wjzhu/OneDrive/PythonCode/MyWork/metradar/testdata/SWAN'

    # filename = 'Z_OTHE_RADAMCR_20230316113000.bin.bz2'
    # filename = 'Z_OTHE_RADAMOSAIC_20230316113000.bin.bz2'
    # filename = 'Z_SWAN_RADAMCC_20230316113000.bin.bz2'
    # filename = 'Z_SWAN_RADAMKDP_20230316113000.bin.bz2'
    # filename = 'Z_SWAN_RADAMZDR_20230316113000.bin.bz2'
    # filename = 'Z_SWAN_RADAMHCA_20230316114000.bin.bz2'
    # filename = 'Z_SWAN_RADAMHCA_20220904110000.bin.bz2'

    if prod_type == 'TDMOSAIC':
        filename = 'Z_OTHE_RADAMOSAIC_' + timestr + '.bin.bz2'
    elif prod_type == 'MCC':
        filename = 'Z_SWAN_RADAMCC_' + timestr + '.bin.bz2'
    elif prod_type == 'MKDP':
        filename = 'Z_SWAN_RADAMKDP_' + timestr + '.bin.bz2'
    elif prod_type == 'MZDR':
        filename = 'Z_SWAN_RADAMZDR_' + timestr + '.bin.bz2'
    elif prod_type == 'HCAMOSAIC':
        filename = 'Z_SWAN_RADAMHCA_' + timestr + '.bin.bz2'
    elif prod_type == 'MCR':
        filename = 'Z_OTHE_RADAMCR_' + timestr + '.bin.bz2'
    elif prod_type == 'TOP':
        filename = 'Z_OTHE_RADAMTOP_' + timestr + '.bin.bz2'
    elif prod_type == 'VIL':
        filename = 'Z_OTHE_RADAMVIL_' + timestr + '.bin.bz2'


    outpath = '/home/wjzhu/OneDrive/PythonCode/MyWork/metradar/pic/SWAN'
    if not os.path.exists(outpath):
        os.makedirs(outpath)
    print('reading file: ',filepath + os.sep + filename)
    if prod_type == 'MCR'  or prod_type == 'TOP' or prod_type == 'VIL':
        lev=0
    else:
        lev = 5
    print('level index: ',lev)
    data = decode_swan(filepath,filename,) # scale=[0.5,-64]
    outname = filename + '_%.1fkm.png'%data.level[lev].values

    slat = 24.5
    nlat = 29.5
    wlon = 103.5
    elon = 109.5
    # slat = float(data.lat.min().data)
    # nlat = float(data.lat.max().data)
    # wlon = float(data.lon.min().data)
    # elon = float(data.lon.max().data)

    newdata = data.data.sel(lat=slice(slat,nlat),lon=slice(wlon,elon))
    # ttdata = newdata.isel(time=0,level=5)
    # plt.show()
    # %%
    # newdata.isel(time=0,level=5).to_netcdf('/Users/wenjianzhu/Downloads/test.nc')
    tstr = filename.split('_')[3].split('.')[0]

    # CC的数据是从0-1，所以要处理一下

    if filename.find('MCR') > 0 :
        draw_latlon(newdata.data[0,lev,:,:],newdata.lat.data,newdata.lon.data,slat,nlat,wlon,elon,
                outpath,outname,tstr,prefix_title='CR',dpi=800,units='dBZ',thred=0,)
    if filename.find('TOP') > 0 :
        draw_latlon(newdata.data[0,lev,:,:],newdata.lat.data,newdata.lon.data,slat,nlat,wlon,elon,
                outpath,outname,tstr,prefix_title='ET',dpi=800,units='km',thred=2,)
    if filename.find('VIL') > 0 :
        draw_latlon(newdata.data[0,lev,:,:],newdata.lat.data,newdata.lon.data,slat,nlat,wlon,elon,
                outpath,outname,tstr,prefix_title='VIL',dpi=800,units='kg/m2',thred=10,)
    elif filename.find('MCC') > 0 :
        draw_latlon(newdata.data[0,lev,:,:]* 100,newdata.lat.data,newdata.lon.data,slat,nlat,wlon,elon,
                outpath,outname,tstr,prefix_title='CAPPI_CC_%.1fkm_'%data.level[lev].values,dpi=800,units='%',thred=0,)
    elif filename.find('MKDP') > 0 :
        draw_latlon(newdata.data[0,lev,:,:],newdata.lat.data,newdata.lon.data,slat,nlat,wlon,elon,
                    outpath,outname,tstr,prefix_title='CAPPI_KDP_%.1fkm_'%data.level[lev].values,dpi=800,units='deg/km',thred=-2)
    elif filename.find('MZDR') > 0: 
        draw_latlon(newdata.data[0,lev,:,:],newdata.lat.data,newdata.lon.data,slat,nlat,wlon,elon,
                outpath,outname,tstr,prefix_title='CAPPI_ZDR_%.1fkm_'%data.level[lev].values,dpi=800,units='dB',thred=-4,)
    elif filename.find('RADAMOSAIC') > 0:
        draw_latlon(newdata.data[0,lev,:,:],newdata.lat.data,newdata.lon.data,slat,nlat,wlon,elon,outpath,
                    outname,tstr,prefix_title='CAPPI_REF_%.1fkm_'%data.level[lev].values,dpi=800,units='dBZ',thred=2,)
    elif filename.find('HCA')  > 0:
        draw_latlon(newdata.data[0,lev,:,:],newdata.lat.data,newdata.lon.data,slat,nlat,wlon,elon,outpath,
                    outname,tstr,prefix_title='CAPPI_HCA_%.1fkm_'%data.level[lev].values,dpi=800,units='cat',thred=0,)
    pass

    '''

    HCA  空=0 地物回波=1 晴空回波=2 干雪=3 湿雪=4 冰晶=5 霰=6  大雨滴=7  小到中雨=8  大雨=9   冰雹=10

    '''



# _*_ coding: utf-8 _*_

'''
记录rose产品的文件结构

'''
import struct


CODE1 = 'B'
CODE2 = 'H'
INT1 = 'B'
INT2 = 'H'
INT4 = 'I'
REAL4 = 'f'
REAL8 = 'd'
SINT1 = 'b'
SINT2 = 'h'
SINT4 = 'i'

# 下面的文件头信息来源于 北京敏视达公司的双偏振天气雷达基数据格式，2016

# 表2-2
GENERIC_HEADER = (
    ('magic_number', INT4),
    ('major_version', INT2),
    ('minor_version', INT2),
    ('generic_type', INT4),
    ('product_type', INT4),
    ('reserved', '16s')
)

# 表2-3
SITE_CONFIG = (
    ('site_code', '8s'),
    ('site_name', '32s'),
    ('lat', REAL4),
    ('lon', REAL4),
    ('ana_height', INT4),
    ('grd_height', INT4),
    ('freq', REAL4),
    ('beamwidth_h', REAL4),
    ('beamwidth_v', REAL4),
    ('rda_version', INT4),
    ('radar_type', INT2),
    ('reserved', '54s')
)
RADAR_TYPE = {
    1: 'SA',
    2: 'SB',
    3: 'SC',
    4: 'SAD',
    5: 'SBD',
    6: 'SCD',
    33: 'CA',
    34: 'CB',
    35: 'CC',
    36: 'CCJ',
    37: 'CD',
    38: 'CAD',
    39: 'CBD',
    40: 'CCD',
    41: 'CCJD',
    42: 'CDD',
    65: 'XA',
    66: 'XAD',
    68: 'XSD'
}

'''
radar_type
1–SA 2–SB
3–SC 4–SAD
5–SBD 6–SCD
33–CA 34–CB
35–CC 36–CCJ
37–CD 38–CAD
39–CBD 40–CCD
41–CCJD 42–CDD
65–XA 66–XAD
68-XSD
'''
# 表2-4
TASK_CONFIG = (
    ('task_name', '32s'),
    ('task_disp', '128s'),
    ('pol_type', INT4),
    ('scan_type', INT4),
    ('pulse_wid', INT4),
    ('scan_stime', INT4),
    ('cut_number', INT4),
    ('hor_noise', REAL4),
    ('ver_noise', REAL4),
    ('hor_calib', REAL4),
    ('ver_calib', REAL4),
    ('hor_noise_t', REAL4),
    ('ver_noise_t', REAL4),
    ('zdr_calib', REAL4),
    ('phi_calib', REAL4),
    ('ldr_calib', REAL4),
    ('reserved', '40s')
)

# 表2-5
SCAN_CONFIG = (
    ('process_mod', INT4),
    ('wave_form', INT4),
    ('prf1', REAL4),
    ('prf2', REAL4),
    ('dias_mode', INT4),
    ('azimuth', REAL4),
    ('elevation', REAL4),
    ('start_angle', REAL4),
    ('end_angle', REAL4),
    ('ang_reso', REAL4),
    ('scan_speed', REAL4),
    ('ref_reso', INT4),
    ('vel_reso', INT4),
    ('max_range1', INT4),
    ('max_range2', INT4),
    ('start_range', INT4),
    ('sample_num1', INT4),
    ('sample_num2', INT4),
    ('phase_mode', INT4),
    ('atmo_loss', REAL4),
    ('nyquist', REAL4),
    ('mom_mask', REAL8),
    ('mom_size_mask', REAL8),
    ('mgk_fil_mask', INT4),
    ('sqi_thr', REAL4),
    ('sig_thr', REAL4),
    ('csr_thr', REAL4),
    ('log_thr', REAL4),
    ('cpa_thr', REAL4),
    ('pmi_thr', REAL4),
    ('dplog_thr', REAL4),
    ('thr_reserved', '4s'),
    ('dbt_mask', INT4),
    ('dbz_mask', INT4),
    ('vel_mask', INT4),
    ('sw_mask', INT4),
    ('dp_mask', INT4),
    ('mask_reserved', '12s'),
    ('scan_sync', INT4),
    ('direction', INT4),
    ('grd_clutter_clsify', INT2),
    ('grd_clutter_filt', INT2),
    ('grd_clutter_filt_ntwidth', INT2),
    ('grd_clutter_filt_window', INT2),
    ('reserved', '72s')
)




# 表3-1 产品头块
PRODUCT_HEADER_BLOCK = (
    ('product_type', INT4),
    ('product_name', '32s'),
    ('product_time_ms', INT4),
    ('scan_start_ms', INT4),
    ('data_start_ms', INT4),
    ('data_end_ms', INT4),
    ('project_type', INT4),
    ('data_type1', INT4),
    ('data_type2', INT4),
    ('reserved', '64s')      
)

# 表 3-4 产品参数块
PRODUCT_DEPENDENT_PARAMETER={
    'PPI':(
        ('ele',REAL4), # 仰角，度
    ), 
    'CR':(
        ('top',INT4,), # 截断顶高，米
        ('bottom',INT4,),# 截断底高，米
    ), 
    'VIL':(
        ('top',INT4,), # 截断顶高，米
        ('bottom',INT4,),# 截断底高，米
    ), 
    'RHI':(
           ('azi',REAL4), # 方位角，度
           ('top',INT4,), #顶高，米
           ('bottom',INT4,), # 底高，米
    ),
    'CAPPI':(
           ('layers',INT4), # 层数
           ('top',INT4,),# 顶高，米
           ('bottom',INT4,),# 底高，米
           ('cappi_fill',INT4,), # 是否填充CAPPI，0-未填充，1-填充
    ),
    'MAX':(
           ('top',INT4,), # 截断顶高，米
           ('bottom',INT4,),# 截断底高，米
    ),
    'ET':(
           ('dbz_ct',REAL4,), # dBZ阈值
    ),
    'VCS':(
           ('start_azi',REAL4), # 起始方位角，度
           ('start_range',INT4,), # 起始距离，米
           ('end_azi',REAL4,), # 结束方位角，度
           ('end_range',INT4,), # 结束距离，米
           ('top',INT4,), # 顶高，米
           ('bottom',INT4,), # 底高，米
    ),
    'LRA':(
           ('top',INT4,),# 顶高，米
           ('bottom',INT4,),# 底高，米
    ),
    'LRM':(
           ('top',INT4,),# 顶高，米
           ('bottom',INT4,),# 底高，米
    ),
    'SRR':(
           ('ele',REAL4), # 仰角，度
           ('range_center',INT4,), # 中心距离，米
           ('azi_center',REAL4,), # 中心方位角，度
           ('side_len',INT4,), # 边长，米
           ('wspd',REAL4,), # 风速，米/秒
           ('wdir',REAL4,), # 风向，度
    ),
    'SRM':( 
           ('ele',REAL4), # 仰角，度
           ('wspd',REAL4,), # 风速，米/秒
           ('wdir',REAL4,), # 风向，度
    ),
    'SWA':(
           ('ele',REAL4),# 仰角，度
           ('range_center',INT4,), # 中心范围，米
           ('azi_center',REAL4,), # 中心方位，度
           ('side_len',INT4,), # 边长，米
    ),
    'WEA':(
           ('range',INT4), # 中心距离，米
           ('azi',REAL4,), # 中心方位角，度
           ('side_len',INT4,), # 边长，米
           ('levels',INT4,), # 层数
    ),
    'OHP':(
           ('base_prod',INT4), # 输入产品类型 产品类型为HSR,CAPPI或QPE，见表3-2
           ('cappi_height',INT4,), # CAPPI高度，米
           ('cappi_fill',INT4,), #是否CAPPI填充
           ('rain_gage_adjust',INT4,),# 是否雨量计修正
    ),
    'THP':(
           ('base_prod',INT4),# 输入产品类型
           ('cappi_height',INT4,),# CAPPI高度，米
           ('cappi_fill',INT4,),#是否CAPPI填充
           ('rain_gage_adjust',INT4,),# 是否雨量计修正
           ('hours',INT4,), # 用户请求的雨量积累的小时数
    ),
    'STP':(
           ('base_prod',INT4),# 输入产品类型
           ('cappi_height',INT4,),# CAPPI高度，米
           ('cappi_fill',INT4,),#是否CAPPI填充
           ('rain_gage_adjust',INT4,),# 是否雨量计修正
    ),
    'USP':(
           ('base_prod',INT4),# 输入产品类型
           ('cappi_height',INT4,),# CAPPI高度，米
           ('cappi_fill',INT4,),#是否CAPPI填充
           ('rain_gage_adjust',INT4,),# 是否雨量计修正
           ('hours',INT4,),# 用户请求的雨量积累的小时数
    ),
    'VAD':(
           ('layers',INT4),# VAD高度的个数
           # 接下来是第1~N层的高度，INT2类型
    ),
    'VWP':(
           ('layers',INT4), # VAD高度的个数
           # 接下来是第1~N层的高度，INT2类型
    ),
    'SHEAR':(
           ('ele',REAL4), # 仰角，度，基于切变数据的仰角
           ('radial_shear',INT4,),# 是否包含径向切变
           ('azi_shear',INT4,), # 是否包含方位切变
           ('ele_shear',INT4,),# 是否包含仰角切变
    ),
    'SWP':(
           ('max_range',INT4), # 最大范围
    ),
    'STI':(
           ('max_range',INT4),# 最大范围
    ),
    'HI':(
           ('max_range',INT4),# 最大范围
    ),
    'M':(
           ('max_range',INT4),# 最大范围
    ),
    'TVS':(
           ('max_range',INT4),# 最大范围
    ),
    'SS':(
           ('max_range',INT4),# 最大范围
    ),
    'GAGE':(
           ('max_range',INT4),# 最大范围
    ),
    'HCL':(
           ('ele',REAL4), # 仰角
    ),
}

# 产品类型列表3-2
PRODUCT_ID_NAME_MAP = {
    1: 'PPI',
    2: 'RHI',
    3: 'CAPPI',
    4: 'MAX',
    6: 'ET',
    8: 'VCS',
    9: 'LRA',
    10: 'LRM',
    13: 'SRR',
    14: 'SRM',
    18: 'CR',
    20: 'WEA',
    23: 'VIL',
    24: 'HSR',
    25: 'OHP',
    26: 'THP',
    27: 'STP',
    28: 'USP',
    31: 'VAD',
    32: 'VWP',
    34: 'SHEAR',
    36: 'SWP',
    37: 'STI',
    38: 'HI',
    39: 'M',
    40: 'TVS',
    41: 'SS',
    48: 'GAGE',
    51: 'HCL',
    52: 'QPE',

}

# 表6-21 ss产品头
SS_HEAD_BLOCK = (
     ('storm_number', INT4), # 风暴个数
)

# 表6-22 ss表结构
SS_TAB = (
    ('storm_id', INT4), # 风暴个数
    ('azi', REAL4), # 方位 度
    ('range', INT4), # 距离 米
    ('base', INT4), # 回波底 米
    ('top', INT4), # 回波顶 米
    ('vil', REAL4), # 垂直积分液态水含量 kg/m**2
    ('mref', REAL4), # 最大反射率 dBZ
    ('hmref', INT4), # 最大反射率高度 米
    
)

# 表6-23 风暴趋势数据
CELL_TREND = (
    ('storm_id', INT4), # 风暴ID
    ('his_vol_num', INT4), # 历史体扫个数,本表03～11项可能会重复
    # CELL_TREND_BASE 3~11项
)
HIS_VOL = (
    ('vol_time', INT4), # 体扫时间，秒
    ('height', INT4), # 高度 米
    ('base_h', INT4), # 回波底 米
    ('top_h', INT4), # 回波顶 米
    ('vil', INT4), # 垂直积分液态水含量 kg/m**2
    ('mref', INT4), # 最大反射率 dBZ
    ('hmref', INT4), # 最大反射率高度 米
    ('ph', INT4), # 冰雹概率 %
    ('psh', INT4), # 强冰雹概率 %
)

# 表6-24 SEGMENT ADAPTATION DATA
SEG_ADAPT = (
    ('REFLECTH1', INT4), # 反射率因子门限
    ('REFLECTH2', INT4), # 反射率因子门限
    ('REFLECTH3', INT4), # 反射率因子门限
    ('REFLECTH4', INT4), # 反射率因子门限
    ('REFLECTH5', INT4), # 反射率因子门限
    ('REFLECTH6', INT4), # 反射率因子门限
    ('REFLECTH7', INT4), # 反射率因子门限
    ('NREFLEVL', INT4), # 反射率因子等级数
    ('NUMAVGBN', INT4),  # 平均库数
    ('SEGRNGMX', INT4), # 段搜索距离
    ('MCOEFCTR', REAL4), # 系数因子
    ('MULTFCTR', REAL4), # 倍数因子
    ('MWGTFCTR', REAL4), # 权重因子

    ('SEGLENTH1', REAL4), # 段长度
    ('SEGLENTH2', REAL4), # 段长度
    ('SEGLENTH3', REAL4), # 段长度
    ('SEGLENTH4', REAL4), # 段长度
    ('SEGLENTH5', REAL4), # 段长度
    ('SEGLENTH6', REAL4), # 段长度
    ('SEGLENTH7', REAL4), # 段长度
    ('DRREFDFF', INT4), # 丢弃反射率因子差
    ('NDROPBIN', INT4),  # 丢弃库数
    ('NUMSEGMX', INT4), # 仰角段数
    ('RADSEGMX', INT4), # 径向段数
    
)

# 表6-25 CENTROIDS  ADAPTATION DATA
CENTROIDS_ADAPT = (
    ('CMPARETH1', INT4), # 风暴组1面积 平方千米
    ('CMPARETH2', INT4), # 风暴组2面积
    ('CMPARETH3', INT4), # 风暴组3面积
    ('CMPARETH4', INT4), # 风暴组4面积
    ('CMPARETH5', INT4), # 风暴组5面积
    ('CMPARETH6', INT4), # 风暴组6面积
    ('CMPARETH7', INT4), # 风暴组7面积
    ('RADIUSTH1', INT4), # 搜索半径1 千米
    ('RADIUSTH2', INT4),  # 搜索半径2
    ('RADIUSTH3', INT4), # 搜索半径3
    ('STMVILMX', REAL4), # 风暴单体VIL最大值 千克/平方米
    ('MXDETSTM', REAL4), # 最大单体个数
    ('OVLAPADJ', REAL4), # 邻近重叠距离 库数

    ('AZMDLTHR', REAL4), # 方位角差门限 度
    ('DEPTHDEL', REAL4), # 删除深度 千米
    ('HORIZDEL', REAL4), # 删除距离 千米
    ('ELVMERGE', REAL4), # 合并仰角 度
    ('HGTMERGE', REAL4), # 合并高度 千米
    ('HRZMERGE', REAL4), # 合并距离 千米
    ('NBRSEGMN', REAL4), # 最少段数
    ('NUMCMPMX', INT4), # 最多组数
    ('MXPOTCMP', INT4),  # 最多可能组数
    ('NUMSTMMX', INT4), # 最多单体数
    
)

# 表6-2
SIT_HEADER_BLOCK = (
    ('storm_number', INT4), # 风暴个数
    ('con_storm_number', INT4), # 连续风暴个数
    ('component_number', INT4), # 构成个数
    ('storm_speed_ave', REAL4), # 平均风暴移动速度 米/秒
    ('storm_dir_ave', REAL4), # 平均风暴移动方向 度
)

# 表6-3
STORM_MOTION_BLOCK = (
    ('azi', REAL4), # 风暴单体到雷达的方位：度
    ('range', INT4), # 风暴到雷达的距离：米
    ('mv_spd', REAL4), # 风暴的移动速度 米/秒
    ('mv_dir', REAL4), # 风暴单体的移动方向 度
    ('forecast_error', INT4),  # 预报错误，米
    ('mean_forecast_error', INT4), # 平均预报错误，米
    
)

# 表6-4 风暴预报或历史信息
STORM_FST_HIS_NUM = ('position_number', INT4), # 随后的位置个数N

STORM_FST_HIS_BLOCK = (
    ('azi', REAL4), # 第1个风暴的方位角，度
    ('range', INT4), # 第1个风暴的距离，米
    ('vol_time_position', INT4),# 第1个风暴的体扫时间，秒
    # 后面是N-1个风暴信息
)

# 表6-5 风暴属性
STORM_PROPERTY = (
    ('id', INT4), # 风暴ID
    ('type', INT4), # 风暴类型 0-连续风暴, 1–新生风暴
    ('vol_num', INT4), # 体扫个数
    ('azi', REAL4), # 方位，度
    ('range', INT4),  # 距离，米
    ('height', INT4), # 高度，米
    ('max_ref', REAL4), # 最大反射率，dBZ
    ('max_ref_height', INT4), # 最大反射率高度，米
    ('vil', REAL4), # VIL值，kg/m**2
    ('num_comp', INT4), # 构成的个数
    ('index_first_comp', INT4), # 第一个构成编号
    ('top_height', INT4), # 风暴顶高，米
    ('index_top', INT4), # 风暴顶的风暴编号
    ('base_height', INT4), # 风暴底高，米
    ('index_base', INT4), # 风暴底风暴编号
)

# 表6-6 风暴构成表
STORM_COMPONENT = (
    ('height', INT4), # 高度，米
    ('max_ref', REAL4), # 最大反射率，dBZ
    ('index_next', INT4), # 下一风暴构成编号
    
)

# 表6-7 风暴追踪适配数据
STORM_TRACK_PARAM = (
    ('def_direc', INT4), # 默认风向，度
    ('def_speed', REAL4),  # 默认风速，米/秒
    ('max_vtime', INT4), # 最大体扫时间间隔，分钟
    ('num_past_vol', INT4), # 历史体扫数
    ('cor_speed', REAL4), # 相关速度，米/秒
    ('min_speed', REAL4), # 最小速度，米/秒
    ('error_allow', INT4), # 允许误差 KM
    ('fst_step', INT4), # 预报间隔，分钟
    ('fst_num', INT4), # 预报个数
    ('error_step', INT4), # 误差间隔，分钟
   
)

# 表6-12 中尺度气旋头信息块
MESO_HEADER_BLOCK = (
    ('storm_number', INT4), # 风暴个数
    ('meso_number', INT4),  # 中尺度气旋个数
    ('feature_number', INT4), # 特征个数
)

# 表6-13 中尺度气旋表块
MESO_TABLE = (
    ('feature_id', INT4), # 中气旋特征ID
    ('storm_id', INT4), # 风暴ID
    ('azi', REAL4), # 方位角 度
    ('range', INT4), # 距离 米
    ('ele', REAL4), # 仰角 度
    ('ave_shear', REAL4), # 平均切变 E-3/S
    ('height', INT4), # 高度 米

    ('azi_diam', INT4), # 方位直径 米
    ('radial_diam', INT4),  # 径向直径 米 
    ('ave_rot_spd', REAL4), # 平均旋转速度 米/秒
    ('max_rot_spd', REAL4), # 最大旋转速度 米/秒
    ('top', INT4), # 顶高 米 
    ('base', INT4), # 底高 米
    ('base_azi', REAL4), # 回波底方位 度
    ('base_range', INT4), # 回波底距离 米
    ('base_ele', REAL4), # 回波底仰角 度
    ('max_t_shear', REAL4), # 最大切向切变 E-3/S
)

# 表6-14 中尺度气旋特征表
MESO_FEATURE_TAB = (
    ('feature_id', INT4), # 中气旋特征ID
    ('storm_id', INT4), # 风暴ID
    ('feature_type', INT4), # 特征类型 1–气旋 2–3D相关切变 3–非相关切变

    ('azi', REAL4), # 方位角 度
    ('range', INT4), # 距离 米
    ('ele', REAL4), # 仰角 度
    ('height', INT4), # 高度 米

    ('azi_diam', INT4), # 方位直径 米
    ('radial_diam', INT4),  # 径向直径 米 
    ('ave_shear', REAL4), # 平均切变 E-3/S
    ('max_shear', REAL4), # 最大切变 E-3/S
    ('ave_rot_spd', REAL4), # 平均旋转速度 米/秒
    ('max_rot_spd', REAL4), # 最大旋转速度 米/秒
    ('top', INT4), # 顶高 米 
    ('base', INT4), # 底高 米
    ('base_azi', REAL4), # 回波底方位 度
    ('base_range', INT4), # 回波底距离 米
    ('base_ele', REAL4), # 回波底仰角 度
   
)

# 表6-15 中尺度适配数据
MESO_ADAPTATION_DATA = (
    ('NPVTHR', INT4), # 特征向量个数门限
    ('FHTHR', REAL4), # 特征高度 千米
    ('HMTHR', REAL4), # 高角动量门限 平方千米/小时

    ('LMTHR', REAL4), # 低角动量门限 平方千米/小时
    ('HSTHR', REAL4), # 高切变门限 1/小时
    ('LSTHR', REAL4), # 低切变门限 1/小时
    ('MRTHR', REAL4), # 直径比率上限

    ('FMRTHR', REAL4), # 远比率上限
    ('NRTHR', REAL4),  # 近比率下限
    ('FNRTHR', REAL4), # 远比率下限
    ('RNGTHR', REAL4), # 距离门限 千米
    ('DISTHR', REAL4), # 最大径向差 千米
    ('AZTHR', REAL4), # 最大方位差 度
   
)


# 表6-17 TVS头信息
TVS_HEADER_BLOCK = (
    ('tvs_number', INT4), # 风暴个数
    ('etvs_numer', INT4),  # 中尺度气旋个数
)

# 表6-18 TVS表信息
TVS_TAB = (
    ('storm_id', INT4), # 风暴ID
    ('type', INT4), # TVS类型 1-TVS  2-ETVS
    ('azi', REAL4), # 方位角 度
    ('range', INT4), # 距离 米
    ('ele', REAL4), # 仰角 度
    ('lldv', REAL4), # 低层速度差值 米/秒
    ('adv', REAL4), # 平均速度差值 米/秒
    ('mxdv', REAL4), # 最大速度差值 米/秒
    ('hmdv', INT4),  # 最大速度差值高度 米
    ('depth', INT4), # 深度 米
    ('base', INT4), # 回波底高 米
    ('top', INT4), # 回波顶高 米 
    ('max_shear', REAL4), # 最大切变 10E-3/S
    ('h_max_shear', INT4), # 最大切变高度 米
   
)

# 表6-19 TVS适配数据
TVS_ADAPTATION_DATA = (
    ('MINREFL', INT4), # 最小反射率
    ('MINPVDV', REAL4), # 最小速度差
    ('MAXPVRNG', REAL4), # 最大模式向量距离

    ('MAXPVHT', REAL4), # 最大模式向量高度
    ('MAXNUMPV', REAL4), # 最大模式向量个数
    ('TH2DDV1', REAL4), # 差分速度门限1
    ('TH2DDV2', REAL4), # 差分速度门限2

    ('TH2DDV3', REAL4), # 差分速度门限3
    ('TH2DDV4', REAL4),  # 差分速度门限4
    ('TH2DDV5', REAL4), # 差分速度门限5
    ('TH2DDV6', REAL4), # 差分速度门限6
    ('MIN1DP2D', REAL4), # 最小模式向量个数
    ('MAXPVRD', REAL4), # 最大模式向量距离

    ('MAXPVAD', REAL4), # 最大模式方位距离
    ('MAX2DAR', REAL4),  # 二维特征最大比率
    ('THCR1', REAL4), # 搜索径向距离1
    ('THCR2', REAL4), # 搜索径向距离2
    ('THCRR', REAL4), # 径向距离门限
    ('MAXNUM2D', REAL4), # 最大二维特征数

    ('MIN2DP3D', REAL4), # 最小二维特征数
    ('MINTVSD', REAL4),  # 最小深度
    ('MINLLDV', REAL4), # 最小低层速度差
    ('MINMTDV', REAL4), # 最小速度差
    ('MAXNUM3D', REAL4), # 最大三维特征个数
    ('MAXNUMTV', REAL4), # 最大TVS个数

    ('MAXNUMET', REAL4), # 最大ETVS个数
    ('MINTVSBH', REAL4),  # TVS最小底高
    ('MINTVSBE', REAL4), # TVS最低仰角
    ('MINADVHT', REAL4), # 最小速度差高度
    ('MAXTSTMD', REAL4), # 最大风暴关联距离

)      

# 冰雹个数
HAIL_NUMBER = (
       ('hail_number', INT4), # 冰雹个数
)

# 表6-9 冰雹表信息
HAIL_TAB = (
    ('hail_id', INT4), # 冰雹ID
    ('azi', REAL4), # 方位角 度
    ('range', INT4), # 距离 米
    ('ph', INT4), # 冰雹概率
    ('psh', INT4), # 极端冰雹概率
    ('hsize', REAL4), # 冰雹大小 ，厘米
    ('rcm_code', INT4), # RCM编码
    
)

# 表6-10 HAIL适配数据
HAIL_ADAPTATION_DATA = (
    ('HT0MSL', INT4), # 0度层高度
    ('HT20MSL', REAL4), # -20度层高度
    ('HKECOF1', REAL4), # 动能系数1

    ('HKECOF2', REAL4), # 动能系数2
    ('HKECOF3', REAL4), # 动能系数3
    ('POSHCOF', REAL4), # 强冰雹概率系数
    ('POSHOFS', REAL4), # 强冰雹概率偏置量

    ('HSCOF', REAL4), # 冰雹尺寸系数
    ('HSEXP', REAL4),  # 冰雹尺寸指数
    ('LLHKEREF', REAL4), # 反射率下限
    ('ULHKEREF', REAL4), # 反射率上限
    ('RCMPRBL', REAL4), # RCM冰雹概率门限
    ('WTCOF', REAL4), # 报警门限系数

    ('MXHLRNG', REAL4), # 冰雹计算最大范围
    ('POHHDTH1', REAL4),  # 冰雹概率1高度差
    ('POHHDTH2', REAL4), # 冰雹概率2高度差
    ('POHHDTH3', REAL4), # 冰雹概率3高度差
    ('POHHDTH4', REAL4), # 冰雹概率4高度差
    ('POHHDTH5', REAL4), # 冰雹概率5高度差

    ('POHHDTH6', REAL4), # 冰雹概率6高度差
    ('POHHDTH7', REAL4),  # 冰雹概率7高度差
    ('POHHDTH8', REAL4), # 冰雹概率8高度差差
    ('POHHDTH9', REAL4), # 冰雹概率9高度差
    ('POHHDTH10', REAL4), # 冰雹概率10高度差
    ('MRPOHTH', REAL4), # 最小反射率门限
    ('RCMPSTV', REAL4), # RCM强冰雹概率
   

)  

# 表2-6 数据类型掩码定义
PRODUCT_DATA_TYPE = {
    1: '滤波前反射率（Total Reflectivity）',
    2: '滤波后反射率(Reflectivity)',
    3: '径向速度(Doppler Velocity)',
    4: '谱宽（Spectrum Width）',
    5: '信号质量指数（Signal Quality Index）',
    6: '杂波相位一致性（Clutter Phase Alignment）',
    7: '差分反射率（Differential Reflectivity）',
    8: '退偏振比（Liner Differential Ratio）',
    9: '协相关系数（Cross Correlation Coefficient）',
    10:'差分相移（Differential Phase）',
    11:'差分相移率（Specific Differential Phase）',
    12:'杂波可能性（Clutter Probability）',
    13:'数据标志，保留',
    14:'双偏振相态分类（Hydro Classification）',
    15:'杂波标志（Clutter Flag）',
    16:'信噪比（Signal Noise Ratio）',
    17:'数据标志，保留',
    18:'数据标志，保留',
    19:'数据标志，保留',
    20:'数据标志，保留',
    21:'数据标志，保留',
    22:'数据标志，保留',
    23:'数据标志，保留',
    24:'数据标志，保留',
    25:'数据标志，保留',
    26:'数据标志，保留',
    27:'数据标志，保留',
    28:'数据标志，保留',
    29:'数据标志，保留',
    30:'数据标志，保留',
    31:'数据标志，保留',
    32:'订正后反射率（Corrected Reflectivity）',
    33:'订正后径向速度(Corrected Doppler Velocity)',
    34:'订正后谱宽（Corrected Spectrum Width）',
    35:'订正后差分反射率(Corrected Differential Reflectivity)',
    72:'回波顶高（Echo Top）',
    73:'垂直积分液态水含量（Vertical Integrated Liquid）',
    75:'累计降水量（Accumulated Precipitation）',
}
# 表4-2 径向头信息块
RADIAL_HEADER=(
    ('data_type', INT4), # 数据类型
    ('scale', INT4), # 数据压缩比
    ('offset', INT4), # 数据偏移
    ('bin_length', INT2), # 数据长度
    ('flags', INT2), # 数据标志
    ('resolution', INT4), # 数据分辨率
    ('start_angle', INT4), # 起始角度
    ('max_range', INT4), # 最大距离
    ('radial_number', INT4), # 径向数
    ('max_value', INT4), # 最大值
    ('range_of_max_value', INT4), # 最大值距禈
    ('azi_of_max_value', REAL4), # 最大值方位
    ('min_value', INT4), # 最小值
    ('range_of_min_value', INT4), # 最小值距离
    ('azi_of_min_value', REAL4), # 最小值方位
    ('reserved', '8s') # 保留

)

# 表4-3 径向数据块信息
RADIAL_DATA=(
    ('start_azi', REAL4), # 起始方位角
    ('anglular_width', REAL4), # 角宽度
    ('num_bins', INT4), # 距离库个数
    ('reserved', '20s'), # 保留数据
    # 后续是第1~N个方位角的数据
)

# 表 4-4 柵格头信息块
GRID_HEADER=(
    ('data_type', INT4), # 数据类型
    ('scale', INT4), # 数据压缩比
    ('offset', INT4), # 数据偏移
    ('bin_length', INT2), # 数据长度
    ('flags', INT2), # 数据标志
    ('row_resolution', INT4), # 横轴分辨率
    ('col_resolution', INT4), # 纵轴分辨率
    ('row_side_len', INT4), # 横轴边长
    ('col_side_len', INT4), # 纵轴边长
    ('max_value', INT4), # 最大值
    ('range_of_max_value', INT4), # 最大值距禿
    ('azi_of_max_value', REAL4), # 最大值方位
    ('min_value', INT4), # 最小值
    ('range_of_min_value', INT4), # 最小值距禿
    ('azi_of_min_value', REAL4), # 最小值方位
    ('reserved', '8s'), # 保留
)

def _structure_size(structure):
    """ Find the size of a structure in bytes. """
    return struct.calcsize('<' + ''.join([i[1] for i in structure]))


def _unpack_structure(string, structure):
    """ Unpack a structure from a string """
    fmt = '<' + ''.join([i[1] for i in structure])  # little-endian
    lst = struct.unpack(fmt, string)
    return dict(zip([i[0] for i in structure], lst))

def _unpack_from_buf( buf, pos, structure):
    """ Unpack a structure from a buffer. """
    size = _structure_size(structure)
    return _unpack_structure(buf[pos:pos + size], structure)


from setuptools import setup, find_packages
import os
# 尝试读取 pyproject.toml
def get_version():
    try:
        # Python 3.11+
        import tomllib
    except ImportError:
        # Python < 3.11，需要安装 tomli
        try:
            import tomli as tomllib
        except ImportError:
            raise RuntimeError("tomli is required to read pyproject.toml in Python < 3.11")

    with open(os.path.join(os.path.dirname(__file__), "pyproject.toml"), "rb") as f:
        pyproject = tomllib.load(f)
    return pyproject["project"]["version"]


setup(
    name="metradar",
    version=get_version(),
    packages=find_packages(),  # 自动发现所有包
)


# python ./setup.py install




# _*_ coding: utf-8 _*_

'''
快速绘制三维雷达组网拼图

'''

# %%
from metradar.graph.draw_comp_mosaic import draw_composite_operational
import os

from metradar.config import CONFIG

# 资源文件路径
RESOURCES_PATH = CONFIG.get('SETTING','RESOURCES_PATH')
COLOR_FILE=RESOURCES_PATH + '/gr2_colors/default_BR_PUP2.pal'

if __name__ == "__main__":

    
    path = '/mnt/e/metradar_test/vpr/mosaic/20230731_daxing'
    outpath = '/mnt/e/metradar_test/vpr/mosaic/20230731_daxing/mosac_pic'

    print(os.listdir(path))
    for file in os.listdir(path):
        if not file.endswith('.nc'):
            continue
        draw_composite_operational(path + os.sep + file,outpath,colorfile=None)
        pass





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 将FMT格式数据直接解析为pyart的radar object格式
# Wenjian Zhu
# 

import os

from pyart.io.nexrad_archive import _find_range_params,_find_scans_to_interp,_interpolate_scan
from pyart.io.common import make_time_unit_str, _test_arguments, prepare_for_read
from pyart.config import FileMetadata, get_fillvalue
from pyart.core.radar import Radar
from .cnrad_level2 import CNRADLevel2File
import warnings
import numpy as np


mapping = {
    "REF": "reflectivity",
    "VEL": "velocity",
    "SW": "spectrum_width",
    "PHI": "differential_phase",
    "ZDR": "differential_reflectivity",
    "RHO": "cross_correlation_ratio",
}

def read_cnrad_fmt(filename, field_names=None, additional_metadata=None,
                        file_field_names=False, exclude_fields=None,
                        include_fields=None, delay_field_loading=False,
                        station=None, scans=None,
                        linear_interp=True, **kwargs):
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata('nexrad_archive', field_names,
                                additional_metadata, file_field_names,
                                exclude_fields, include_fields)

    # open the file and retrieve scan information
    nfile = CNRADLevel2File(prepare_for_read(filename))
    scan_info = nfile.scan_info(scans)

    # time
    time = filemetadata('time')
    time_start, _time = nfile.get_times(scans)
    time['data'] = _time
    time['units'] = make_time_unit_str(time_start)

    # range
    _range = filemetadata('range')
    first_gate, gate_spacing, last_gate = _find_range_params(
        scan_info, filemetadata)
    _range['data'] = np.arange(first_gate, last_gate, gate_spacing, 'float32')
    _range['meters_to_center_of_first_gate'] = float(first_gate)
    _range['meters_between_gates'] = float(gate_spacing)

    # metadata
    metadata = filemetadata('metadata')
    metadata['original_container'] = 'CINRAD_FMT'
    vcp_pattern = nfile.get_vcp_pattern()
    if vcp_pattern is not None:
        metadata['vcp_pattern'] = vcp_pattern
    if 'icao' in nfile.volume_header.keys():
        metadata['instrument_name'] = nfile.volume_header['icao'].decode()

    # scan_type
    scan_type = 'ppi'

    # latitude, longitude, altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')

    # if nfile._msg_type == '1' and station is not None:
    #     lat, lon, alt = get_nexrad_location(station)
    # elif 'icao' in nfile.volume_header.keys() and nfile.volume_header['icao'].decode()[0] == 'T':
    #     lat, lon, alt = get_nexrad_location(
    #         nfile.volume_header['icao'].decode())
    # else:
    #     lat, lon, alt = nfile.location()

    lat, lon, alt = nfile.location()

    latitude['data'] = np.array([lat], dtype='float64')
    longitude['data'] = np.array([lon], dtype='float64')
    altitude['data'] = np.array([alt], dtype='float64')

    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index
    # sweep_end_ray_index
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    if scans is None:
        nsweeps = int(nfile.nscans)
    else:
        nsweeps = len(scans)
    sweep_number['data'] = np.arange(nsweeps, dtype='int32')
    sweep_mode['data'] = np.array(
        nsweeps * ['azimuth_surveillance'], dtype='S')

    rays_per_scan = [s['nrays'] for s in scan_info]
    sweep_end_ray_index['data'] = np.cumsum(rays_per_scan, dtype='int32') - 1

    rays_per_scan.insert(0, 0)
    sweep_start_ray_index['data'] = np.cumsum(
        rays_per_scan[:-1], dtype='int32')

    # azimuth, elevation, fixed_angle
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')
    fixed_angle = filemetadata('fixed_angle')
    azimuth['data'] = nfile.get_azimuth_angles(scans)
    elevation['data'] = nfile.get_elevation_angles(scans).astype('float32')
    fixed_agl = []
    for i in nfile.get_target_angles(scans):
        if i > 180:
            i = i - 360.
            warnings.warn("Fixed_angle(s) greater than 180 degrees present."
                          + " Assuming angle to be negative so subtrating 360",
                          UserWarning)
        else:
            i = i
        fixed_agl.append(i)
    fixed_angles = np.array(fixed_agl, dtype='float32')
    fixed_angle['data'] = fixed_angles

    # fields
    max_ngates = len(_range['data'])
    available_moments = set([m for scan in scan_info for m in scan['moments']])
    interpolate = _find_scans_to_interp(
        scan_info, first_gate, gate_spacing, filemetadata)

    fields = {}
    for moment in available_moments:
        field_name = filemetadata.get_field_name(moment)
        if field_name is None:
            continue
        dic = filemetadata(field_name)
        dic['_FillValue'] = get_fillvalue()
        if delay_field_loading and moment not in interpolate:
            # dic = LazyLoadDict(dic)
            # data_call = _NEXRADLevel2StagedField(
            #     nfile, moment, max_ngates, scans)
            # dic.set_lazy('data', data_call)
            pass
        else:
            mdata = nfile.get_data(moment, max_ngates, scans=scans)
            if moment in interpolate:
                interp_scans = interpolate[moment]
                warnings.warn(
                    "Gate spacing is not constant, interpolating data in " +
                    "scans %s for moment %s." % (interp_scans, moment),
                    UserWarning)
                for scan in interp_scans:
                    idx = scan_info[scan]['moments'].index(moment)
                    moment_ngates = scan_info[scan]['ngates'][idx]
                    start = sweep_start_ray_index['data'][scan]
                    end = sweep_end_ray_index['data'][scan]
                    if interpolate['multiplier'] == '4':
                        multiplier = '4'
                    else:
                        multiplier = '2'
                    _interpolate_scan(mdata, start, end, moment_ngates,
                                      multiplier, linear_interp)
            dic['data'] = mdata
        fields[field_name] = dic

    # instrument_parameters
    nyquist_velocity = filemetadata('nyquist_velocity')
    unambiguous_range = filemetadata('unambiguous_range')
    nyquist_velocity['data'] = nfile.get_nyquist_vel(scans).astype('float32')
    unambiguous_range['data'] = (
        nfile.get_unambigous_range(scans).astype('float32'))

    radar_beam_width_h = filemetadata('radar_beam_width_h')
    radar_beam_width_v = filemetadata('radar_beam_width_v')
    radar_beam_width_h['data'] = nfile.get_beam_width_h()
    radar_beam_width_v['data'] = nfile.get_beam_width_v()
    
    radar_atenna_gain = filemetadata('radar_antenna_gain')
    radar_receiver_bandwidth = filemetadata('radar_receiver_bandwidth')
    radar_atenna_gain['data'] = nfile.get_antenna_gain()
    radar_receiver_bandwidth['data'] = nfile.get_receiver_bandwidth()
    

    instrument_parameters = {'unambiguous_range': unambiguous_range,
                             'nyquist_velocity': nyquist_velocity, 
                             'radar_beam_width_h': radar_beam_width_h,
                             'radar_beam_width_v': radar_beam_width_v,
                             'radar_antenna_gain_h': radar_atenna_gain,
                             'radar_antenna_gain_v': radar_atenna_gain,
                             'radar_receiver_bandwidth': radar_receiver_bandwidth,
                             }

    nfile.close()
    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)

if __name__ == "__main__":
    pass
    
    filepath = './testdata/FMT/'
    filename = 'Z_RADR_I_Z9417_20210910154708_O_DOR_SAD_CAP_FMT.bin'

    radar = read_cnrad_fmt(filepath + os.sep + filename)

    print(list(radar.fields.keys()))
    pass